import calendar
from datetime import datetime, timedelta

from django.utils.timezone import now


class Helper:
    @property
    def thirty_days_ago(self) -> datetime:
        return self.today - timedelta(days=30)

    @property
    def sixty_days_ago(self) -> datetime:
        return self.today - timedelta(days=60)

    @property
    def ninety_days_ago(self) -> datetime:
        return self.today - timedelta(days=90)

    @property
    def today(self) -> datetime:
        return now()

    @property
    def tomorrow(self) -> datetime:
        return self.today + timedelta(days=1)

    @property
    def yesterday(self) -> datetime:
        return self.today - timedelta(days=1)

    @property
    def start_of_current_week(self) -> datetime:
        return self.today - timedelta(days=self.today.weekday())

    @property
    def end_of_current_week(self) -> datetime:
        return self.start_of_week + timedelta(days=6)

    @property
    def start_of_last_week(self) -> datetime:
        return self.start_of_week - timedelta(days=6)

    @property
    def end_of_last_week(self) -> datetime:
        return self.end_of_week - timedelta(days=6)

    @property
    def start_of_current_month(self) -> datetime:
        return self.today.replace(day=1)

    @property
    def end_of_current_month(self) -> datetime:
        _, last_day = calendar.monthrange(self.today.year, self.today.month)
        return self.today.replace(day=last_day)

    @property
    def start_of_last_month(self) -> datetime:
        return self._add_months(self.today, -1).replace(day=1)

    @property
    def end_of_last_month(self) -> datetime:
        last_month = self._add_months(self.today, -1)
        _, last_day = calendar.monthrange(last_month.year, last_month.month)

        return last_month.replace(day=last_day)

    @property
    def start_of_current_year(self) -> datetime:
        return self.today.replace(month=1, day=1)

    @property
    def end_of_current_year(self) -> datetime:
        return self.today.replace(month=12, day=31)

    @property
    def start_of_last_year(self) -> datetime:
        return self.today.replace(year=self.today.year - 1, month=1, day=1)

    @property
    def end_of_last_year(self) -> datetime:
        return self.today.replace(year=self.today.year - 1, month=12, day=31)

    @staticmethod
    def _add_months(datetime_, months):
        month = datetime_.month - 1 + months
        year = datetime_.year + month // 12
        month = month % 12 + 1

        day = min(datetime_.day, calendar.monthrange(year, month)[1])

        return datetime_.replace(year=year, month=month, day=day)


from __future__ import annotations

from typing import TYPE_CHECKING

from django.http import JsonResponse
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.decorators import login_required

from django_spire.notification.app.models import AppNotification
from django_spire.history.viewed.models import Viewed

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


@login_required()
def check_new_notifications_ajax_view(request: WSGIRequest) -> JsonResponse:
    notifications = (
        AppNotification.objects
        .active()
        .is_sent()
        .exclude_viewed_by_user(request.user)
    )

    return JsonResponse({
        'status': 200,
        'has_new_notifications': bool(notifications)
    })


@login_required()
def set_notifications_as_viewed_ajax(request: WSGIRequest) -> JsonResponse:
    notification_list = (
        AppNotification.objects
        .active()
        .is_sent()
        .by_user(request.user)
        .exclude_viewed_by_user(request.user)
    )
    ctype = ContentType.objects.get_for_model(AppNotification)
    viewed_model_objects = [
        Viewed(
            user=request.user,
            object_id=notification.id,
            content_type=ctype
        )
        for notification in notification_list
    ]

    Viewed.objects.bulk_create(viewed_model_objects)

    return JsonResponse({'status': 200, 'message': 'Notifications succesfully marked as viewed.'})



from __future__ import annotations

import marko

from django import template
from django.utils.safestring import mark_safe


register = template.Library()


@register.filter
def render_markdown(value: str) -> str:
    if not value:
        return ''

    html = marko.convert(value)

    return mark_safe(html)



from __future__ import annotations

import json

from typing import TYPE_CHECKING

from django.http import JsonResponse

from django_spire.core.shortcuts import get_object_or_null_obj
from django_spire.core.decorators import valid_ajax_request_required
from django_spire.knowledge.collection.models import Collection
from django_spire.knowledge.entry.models import Entry

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


@valid_ajax_request_required
def reorder_view(request: WSGIRequest) -> JsonResponse:
    body_data = json.loads(request.body.decode('utf-8'))

    entry_id = body_data.get('entry_id', 0)
    entry = get_object_or_null_obj(Entry, pk=entry_id)

    if entry.id is None:
        return JsonResponse({'type': 'error', 'message': 'Entry not found.'})

    order = body_data.get('order', None)

    if order is None:
        return JsonResponse({'type': 'error', 'message': 'Order not found.'})

    collection_id = body_data.get('collection_id', None)
    collection = get_object_or_null_obj(Collection, pk=collection_id)

    if collection.id is None:
        return JsonResponse({'type': 'error', 'message': 'Collection not found.'})

    entry.collection = collection
    entry.save()

    entry.ordering_services.processor.move_to_position(
        destination_objects=(
            collection.entries
            .has_current_version()
            .user_has_access(user=request.user)
            .active()
        ),
        position=order,
    )

    return JsonResponse({'type': 'success', 'message': 'Order reordered successfully'})


@valid_ajax_request_required
def update_files_view(request: WSGIRequest) -> JsonResponse:
    return JsonResponse({'files_json': Entry.services.tool.get_files_to_convert_json()})



from __future__ import annotations

import pytest

from dataclasses import fields

from django.test import TestCase

from django_spire.theme.enums import ThemeFamily, ThemeMode
from django_spire.theme.models import Theme


class ThemeModelTests(TestCase):
    def test_dataclass_fields(self) -> None:
        theme_fields = {f.name: f.type for f in fields(Theme)}

        assert set(theme_fields.keys()) == {'family', 'mode'}
        assert theme_fields['family'] == 'ThemeFamily'
        assert theme_fields['mode'] == 'ThemeMode'

    def test_class_variables(self) -> None:
        assert Theme.DEFAULT_FAMILY == ThemeFamily.DEFAULT
        assert Theme.DEFAULT_MODE == ThemeMode.LIGHT
        assert Theme.SEPARATOR == '-'

        for family in ThemeFamily:
            assert family in Theme.FAMILY_DISPLAY_NAMES

    def test_theme_initialization_with_enums(self) -> None:
        theme = Theme(family=ThemeFamily.GRUVBOX, mode=ThemeMode.DARK)
        assert theme.family == ThemeFamily.GRUVBOX
        assert theme.mode == ThemeMode.DARK

    def test_theme_initialization_with_strings(self) -> None:
        theme = Theme(family='gruvbox', mode='dark')
        assert theme.family == ThemeFamily.GRUVBOX
        assert theme.mode == ThemeMode.DARK

    def test_theme_initialization_invalid_family(self) -> None:
        with pytest.raises(ValueError) as ctx:
            Theme(family='invalid-family', mode='dark')

        assert 'Invalid theme family' in str(ctx.value)

    def test_theme_initialization_invalid_mode(self) -> None:
        with pytest.raises(ValueError) as ctx:
            Theme(family='gruvbox', mode='invalid-mode')

        assert 'Invalid theme mode' in str(ctx.value)

    def test_from_string_valid(self) -> None:
        cases = [
            ('default-light', ThemeFamily.DEFAULT, ThemeMode.LIGHT),
            ('gruvbox-dark', ThemeFamily.GRUVBOX, ThemeMode.DARK),
            ('one-dark-light', ThemeFamily.ONE_DARK, ThemeMode.LIGHT),
            ('tokyo-night-dark', ThemeFamily.TOKYO_NIGHT, ThemeMode.DARK),
        ]

        for string, family, mode in cases:
            with self.subTest(string=string):
                theme = Theme.from_string(string)
                assert theme.family == family
                assert theme.mode == mode

    def test_from_string_empty_with_default(self) -> None:
        default = Theme(family=ThemeFamily.GRUVBOX, mode=ThemeMode.DARK)
        theme = Theme.from_string('', default=default)
        assert theme == default

    def test_from_string_empty_without_default(self) -> None:
        theme = Theme.from_string('')
        assert theme.family == Theme.DEFAULT_FAMILY
        assert theme.mode == Theme.DEFAULT_MODE

    def test_from_string_invalid_with_default(self) -> None:
        default = Theme(family=ThemeFamily.GRUVBOX, mode=ThemeMode.DARK)
        theme = Theme.from_string('invalid', default=default)
        assert theme == default

    def test_from_string_invalid_without_default(self) -> None:
        with pytest.raises(ValueError):
            Theme.from_string('invalid')

    def test_get_available(self) -> None:
        available = Theme.get_available()

        count = len(ThemeFamily) * len(ThemeMode)
        assert len(available) == count

        for family in ThemeFamily:
            for mode in ThemeMode:
                theme = Theme(family=family, mode=mode)
                assert theme in available

    def test_get_default(self) -> None:
        default = Theme.get_default()
        assert default.family == Theme.DEFAULT_FAMILY
        assert default.mode == Theme.DEFAULT_MODE

    def test_display_property(self) -> None:
        theme = Theme(family=ThemeFamily.GRUVBOX, mode=ThemeMode.DARK)
        assert theme.display == 'Gruvbox - Dark'

        theme = Theme(family=ThemeFamily.ONE_DARK, mode=ThemeMode.LIGHT)
        assert theme.display == 'One Dark Pro - Light'

    def test_family_display_property(self) -> None:
        theme = Theme(family=ThemeFamily.ONE_DARK, mode=ThemeMode.DARK)
        assert theme.family_display == 'One Dark Pro'

    def test_is_dark_property(self) -> None:
        dark = Theme(family=ThemeFamily.GRUVBOX, mode=ThemeMode.DARK)
        assert dark.is_dark

        light = Theme(family=ThemeFamily.GRUVBOX, mode=ThemeMode.LIGHT)
        assert not light.is_dark

    def test_stylesheet_property(self) -> None:
        cases = [
            (ThemeFamily.DEFAULT, ThemeMode.LIGHT, 'django_spire/css/themes/default/app-light.css'),
            (ThemeFamily.GRUVBOX, ThemeMode.DARK, 'django_spire/css/themes/gruvbox/app-dark.css'),
            (ThemeFamily.ONE_DARK, ThemeMode.LIGHT, 'django_spire/css/themes/one-dark/app-light.css'),
            (ThemeFamily.TOKYO_NIGHT, ThemeMode.DARK, 'django_spire/css/themes/tokyo-night/app-dark.css'),
        ]

        for family, mode, path in cases:
            with self.subTest(family=family, mode=mode):
                theme = Theme(family=family, mode=mode)
                assert theme.stylesheet == path

    def test_value_property(self) -> None:
        cases = [
            (ThemeFamily.DEFAULT, ThemeMode.LIGHT, 'default-light'),
            (ThemeFamily.GRUVBOX, ThemeMode.DARK, 'gruvbox-dark'),
            (ThemeFamily.ONE_DARK, ThemeMode.LIGHT, 'one-dark-light'),
            (ThemeFamily.TOKYO_NIGHT, ThemeMode.DARK, 'tokyo-night-dark'),
        ]

        for family, mode, value in cases:
            with self.subTest(family=family, mode=mode):
                theme = Theme(family=family, mode=mode)
                assert theme.value == value

    def test_to_dict(self) -> None:
        theme = Theme(family=ThemeFamily.GRUVBOX, mode=ThemeMode.DARK)
        result = theme.to_dict()

        keys = {
            'display',
            'family',
            'family_display',
            'full',
            'is_dark',
            'mode',
            'stylesheet'
        }

        assert set(result.keys()) == keys

        assert result['display'] == 'Gruvbox - Dark'
        assert result['family'] == 'gruvbox'
        assert result['family_display'] == 'Gruvbox'
        assert result['full'] == 'gruvbox-dark'
        assert result['is_dark']
        assert result['mode'] == 'dark'
        assert result['stylesheet'] == 'django_spire/css/themes/gruvbox/app-dark.css'

    def test_theme_immutability(self) -> None:
        theme = Theme(family=ThemeFamily.GRUVBOX, mode=ThemeMode.DARK)

        with pytest.raises(AttributeError):
            theme.family = ThemeFamily.DEFAULT

        with pytest.raises(AttributeError):
            theme.mode = ThemeMode.LIGHT



from __future__ import annotations

from django_spire.ai.sms.intel import SmsIntel
from django_spire.core.tests.test_cases import BaseTestCase


class SmsIntelTests(BaseTestCase):
    def test_sms_intel_creation(self) -> None:
        intel = SmsIntel(body='Test message')

        assert intel.body == 'Test message'

    def test_sms_intel_empty_body(self) -> None:
        intel = SmsIntel(body='')

        assert intel.body == ''

    def test_sms_intel_long_body(self) -> None:
        long_body = 'A' * 1000
        intel = SmsIntel(body=long_body)

        assert intel.body == long_body
        assert len(intel.body) == 1000

    def test_sms_intel_unicode_body(self) -> None:
        unicode_body = 'Hello 世界 🌍'
        intel = SmsIntel(body=unicode_body)

        assert intel.body == unicode_body

    def test_sms_intel_model_dump(self) -> None:
        intel = SmsIntel(body='Test')
        dump = intel.model_dump()

        assert 'body' in dump
        assert dump['body'] == 'Test'

    def test_sms_intel_special_characters(self) -> None:
        special_body = 'Line 1\nLine 2\tTabbed'
        intel = SmsIntel(body=special_body)

        assert intel.body == special_body



from __future__ import annotations

from django.test import TestCase

from django_spire.ai.sms.models import SmsConversation, SmsMessage


class SmsConversationModelTests(TestCase):
    def test_conversation_creation(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )

        assert conversation.phone_number == '+15551234567'
        assert conversation.messages.count() == 0
        assert conversation.is_empty

    def test_conversation_str(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )

        assert '+15551234567' in str(conversation)
        assert 'SMS Conversation' in str(conversation)

    def test_add_message(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )

        message = conversation.add_message(body='Hello', is_inbound=True, twilio_sid='')

        assert message.body == 'Hello'
        assert message.is_inbound

        message = conversation.add_message(body='Hi there', is_inbound=False, twilio_sid='')

        assert message.body == 'Hi there'
        assert not message.is_inbound

        assert conversation.messages.count() == 2
        assert not conversation.is_empty

    def test_add_message_updates_last_message_datetime(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )

        initial_datetime = conversation.last_message_datetime

        conversation.add_message(body='Test', is_inbound=True, twilio_sid='')
        conversation.refresh_from_db()

        assert conversation.last_message_datetime >= initial_datetime

    def test_generate_message_history(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )

        conversation.add_message(body='User message 1', is_inbound=True, twilio_sid='')
        conversation.add_message(body='Assistant response 1', is_inbound=False, twilio_sid='')
        conversation.add_message(body='User message 2', is_inbound=True, twilio_sid='')

        message_history = conversation.generate_message_history(
            message_count=20,
            exclude_last_message=True
        )

        assert message_history is not None

    def test_generate_message_history_includes_last_message(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )

        conversation.add_message(body='User message 1', is_inbound=True, twilio_sid='')
        conversation.add_message(body='User message 2', is_inbound=True, twilio_sid='')

        message_history = conversation.generate_message_history(
            message_count=20,
            exclude_last_message=False
        )

        assert message_history is not None

    def test_is_empty_property(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )

        assert conversation.is_empty

        conversation.add_message(body='Test', is_inbound=True, twilio_sid='')

        assert not conversation.is_empty


class SmsMessageModelTests(TestCase):
    def test_message_creation(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )
        message = SmsMessage.objects.create(
            conversation=conversation,
            body='Test message',
            is_inbound=True,
            twilio_sid='SM123'
        )

        assert message.body == 'Test message'
        assert message.is_inbound
        assert message.twilio_sid == 'SM123'

    def test_message_str_short(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )
        message = SmsMessage.objects.create(
            conversation=conversation,
            body='Short message',
            is_inbound=True
        )

        assert 'Inbound' in str(message)
        assert 'Short message' in str(message)

    def test_message_str_long(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )
        long_body = 'A' * 100
        message = SmsMessage.objects.create(
            conversation=conversation,
            body=long_body,
            is_inbound=True
        )

        assert '...' in str(message)

    def test_direction_property_inbound(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )
        message = SmsMessage.objects.create(
            conversation=conversation,
            body='Test',
            is_inbound=True
        )

        assert message.direction == 'Inbound'

    def test_direction_property_outbound(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )
        message = SmsMessage.objects.create(
            conversation=conversation,
            body='Test',
            is_inbound=False
        )

        assert message.direction == 'Outbound'

    def test_is_outbound_property(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )

        inbound_message = SmsMessage.objects.create(
            conversation=conversation,
            body='Test',
            is_inbound=True
        )

        outbound_message = SmsMessage.objects.create(
            conversation=conversation,
            body='Test',
            is_inbound=False
        )

        assert not inbound_message.is_outbound
        assert outbound_message.is_outbound

    def test_role_property_inbound(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )
        message = SmsMessage.objects.create(
            conversation=conversation,
            body='Test',
            is_inbound=True
        )

        assert message.role == 'user'

    def test_role_property_outbound(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )
        message = SmsMessage.objects.create(
            conversation=conversation,
            body='Test',
            is_inbound=False
        )

        assert message.role == 'assistant'

    def test_message_is_processed_default(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )
        message = SmsMessage.objects.create(
            conversation=conversation,
            body='Test',
            is_inbound=True
        )

        assert message.is_processed is False


class SmsQuerySetTests(TestCase):
    def test_by_phone_number(self) -> None:
        SmsConversation.objects.create(phone_number='+15551234567')
        SmsConversation.objects.create(phone_number='+15559876543')

        result = SmsConversation.objects.by_phone_number('+15551234567')

        assert result.count() == 1
        assert result.first().phone_number == '+15551234567'

    def test_newest_by_count(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )

        for i in range(25):
            conversation.add_message(body=f'Message {i}', is_inbound=True, twilio_sid='')

        messages = conversation.messages.newest_by_count(20)

        assert len(messages) == 20

    def test_newest_by_count_reversed(self) -> None:
        conversation = SmsConversation.objects.create(
            phone_number='+15551234567'
        )

        for i in range(5):
            conversation.add_message(body=f'Message {i}', is_inbound=True, twilio_sid='')

        messages = conversation.messages.newest_by_count_reversed(5)

        assert len(messages) == 5



from __future__ import annotations

from django_spire.auth.group.models import AuthGroup
from django_spire.auth.group.querysets import GroupQuerySet
from django_spire.core.tests.test_cases import BaseTestCase


class GroupQuerySetTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.group_c = AuthGroup.objects.create(name='Charlie Group')
        self.group_a = AuthGroup.objects.create(name='Alpha Group')
        self.group_b = AuthGroup.objects.create(name='Beta Group')

    def test_active_returns_ordered_by_name(self) -> None:
        queryset = AuthGroup.objects.all()
        queryset.__class__ = GroupQuerySet
        result = queryset.active()
        names = list(result.values_list('name', flat=True))
        assert names == ['Alpha Group', 'Beta Group', 'Charlie Group']

    def test_active_returns_all_groups(self) -> None:
        queryset = AuthGroup.objects.all()
        queryset.__class__ = GroupQuerySet
        result = queryset.active()
        assert result.count() == 3

    def test_active_empty_queryset(self) -> None:
        AuthGroup.objects.all().delete()
        queryset = AuthGroup.objects.all()
        queryset.__class__ = GroupQuerySet
        result = queryset.active()
        assert result.count() == 0

    def test_active_single_group(self) -> None:
        AuthGroup.objects.all().delete()
        AuthGroup.objects.create(name='Single Group')
        queryset = AuthGroup.objects.all()
        queryset.__class__ = GroupQuerySet
        result = queryset.active()
        assert result.count() == 1

    def test_active_ordering_with_numbers(self) -> None:
        group1 = AuthGroup.objects.create(name='AAA First')
        group2 = AuthGroup.objects.create(name='BBB Second')
        group3 = AuthGroup.objects.create(name='CCC Third')
        queryset = AuthGroup.objects.filter(pk__in=[group1.pk, group2.pk, group3.pk])
        queryset.__class__ = GroupQuerySet
        result = queryset.active()
        names = list(result.values_list('name', flat=True))
        assert names == ['AAA First', 'BBB Second', 'CCC Third']

    def test_active_ordering_with_special_characters(self) -> None:
        AuthGroup.objects.create(name='!Special')
        AuthGroup.objects.create(name='@Another')
        queryset = AuthGroup.objects.all()
        queryset.__class__ = GroupQuerySet
        result = queryset.active()
        assert result.count() >= 2

    def test_active_ordering_case_sensitive(self) -> None:
        AuthGroup.objects.all().delete()
        AuthGroup.objects.create(name='alpha')
        AuthGroup.objects.create(name='Alpha')
        AuthGroup.objects.create(name='ALPHA')
        queryset = AuthGroup.objects.all()
        queryset.__class__ = GroupQuerySet
        result = queryset.active()
        names = list(result.values_list('name', flat=True))
        assert len(names) == 3

    def test_active_ordering_with_unicode(self) -> None:
        AuthGroup.objects.create(name='Ñame')
        AuthGroup.objects.create(name='Ääkkönen')
        queryset = AuthGroup.objects.all()
        queryset.__class__ = GroupQuerySet
        result = queryset.active()
        assert result.count() >= 2

    def test_active_is_chainable(self) -> None:
        queryset = AuthGroup.objects.all()
        queryset.__class__ = GroupQuerySet
        result = queryset.active().filter(name__contains='Group')
        assert result.count() == 3

    def test_active_with_prefetch(self) -> None:
        queryset = AuthGroup.objects.all().prefetch_related('permissions')
        queryset.__class__ = GroupQuerySet
        result = queryset.active()
        assert result.count() == 3

    def test_active_returns_queryset(self) -> None:
        queryset = AuthGroup.objects.all()
        queryset.__class__ = GroupQuerySet
        result = queryset.active()
        assert hasattr(result, 'filter')
        assert hasattr(result, 'exclude')



from __future__ import annotations

import json

from django_spire.knowledge.entry.version.block.choices import BlockTypeChoices
from django_spire.knowledge.entry.version.block.data.text_data import TextEditorBlockData
from django_spire.knowledge.entry.version.block.models import EntryVersionBlock
from django_spire.knowledge.entry.version.tests.factories import create_test_entry_version


def create_test_block_form_data(**kwargs) -> dict:
    data = {
        'id': 'ABC123',
        'order': 0,
        'type': BlockTypeChoices.TEXT,
        'data': {'text': 'test text'},
        'tunes': {}
    }
    data.update(kwargs)
    return data


def create_test_version_block(**kwargs) -> EntryVersionBlock:
    data = {
        'version': kwargs.pop('version', None) or create_test_entry_version(),
        'order': 1,
        'type': BlockTypeChoices.TEXT,
        '_block_data': json.dumps({}),
        '_text_data': ''
    }
    data.update(kwargs)
    version_block = EntryVersionBlock.objects.create(**data)
    version_block.editor_js_block_data = TextEditorBlockData(text='')
    version_block.save()

    return version_block



from __future__ import annotations

import logging
import marko
import re

from collections import defaultdict
from typing import TYPE_CHECKING

from bs4 import BeautifulSoup
from django.core.files.storage import default_storage
from markitdown.converters import HtmlConverter
from marko.block import Heading, List, ListItem, Paragraph, BlankLine

from django_spire.knowledge.entry.version.block import models
from django_spire.knowledge.entry.version.block.choices import BlockTypeChoices
from django_spire.knowledge.entry.version.block.data.list.choices import ListEditorBlockDataStyle
from django_spire.knowledge.entry.version.converters.converter import BaseConverter

if TYPE_CHECKING:
    from marko.element import Element
    from django_spire.knowledge.entry.version.models import EntryVersion
    from django_spire.file.models import File
    from marko.block import BlockElement


MARKO_BLOCK_TYPE_TO_BLOCK_CHOICES = defaultdict(
    lambda:BlockTypeChoices.TEXT,
    {
        Paragraph: models.BlockTypeChoices.TEXT,
        BlankLine: models.BlockTypeChoices.TEXT,
        Heading: models.BlockTypeChoices.HEADING,
        List: models.BlockTypeChoices.LIST,
    }
)


class MarkdownConverter(BaseConverter):
    """Converts Markdown content to a list of EntryVersionBlocks using Marko.

    For more info on Marko:
    https://marko-py.readthedocs.io/en/latest/api.html#marko.block.BlockElement
    """

    html_converter = HtmlConverter()

    def __init__(self, entry_version: EntryVersion):
        super().__init__(entry_version)
        self._order = 0

    def convert_file_to_blocks(self, file: File) -> list[models.EntryVersionBlock]:
        with default_storage.open(file.file.name, 'r') as f:
            return self.convert_markdown_to_blocks(f.read())

    def convert_markdown_to_blocks(
        self,
        markdown_content: str
    ) -> list[models.EntryVersionBlock]:
        marko_blocks = marko.parse(markdown_content).children

        return [
            models.EntryVersionBlock.services.factory.create_validated_block(
                entry_version=self.entry_version,
                type=MARKO_BLOCK_TYPE_TO_BLOCK_CHOICES[marko_block.__class__],
                data=self._marko_block_to_editor_block_data_dict(marko_block),
                order=order,
                tunes={}
            )
            for order, marko_block in enumerate(marko_blocks)
        ]

    @classmethod
    def html_to_markdown(cls, html_content: str) -> str:
        return cls.html_converter.convert_string(html_content).markdown

    @classmethod
    def _remove_outer_html_tags(
            cls,
            html_content: str,
            tag_name: str | None = None,
    ) -> str:
        bs = BeautifulSoup(html_content, 'html.parser')

        if tag_name:
            # Only target specific tag given for removal
            html_element = bs.find(tag_name)

            if html_element:
                html_content = html_element.decode_contents()
        else:
            # Remove any outer tag found
            html_element = bs.find()

            if html_element:
                html_content = html_element.decode_contents()

        return html_content

    @staticmethod
    def _try_parse_content_as_checklist_item(content: str) -> tuple[bool, bool | None, str]:
        empty_checkbox_pattern = re.compile(r'\[ ?\]\s*')
        checked_checkbox_pattern = re.compile(r'\[[xX]\]\s*')

        is_checked = None
        has_empty_checkbox = re.match(empty_checkbox_pattern, content)

        if has_empty_checkbox:
            content = re.sub(empty_checkbox_pattern, '', content)
            is_checked = False

        has_checked_checkbox = re.match(checked_checkbox_pattern, content)
        if has_checked_checkbox:
            content = re.sub(checked_checkbox_pattern, '', content)
            is_checked = True

        is_checklist_item = is_checked is not None

        return (
            is_checklist_item,
            is_checked,
            content
        )

    @classmethod
    def _marko_list_item_block_to_editor_block_data_dict(
        cls,
        marko_list_item_block: ListItem,
        parent_marko_list_block: List | None = None
    ):
        list_item_editor_block_data_kwargs = {
            'items': [],
            'content': '',
            'meta': {},
        }

        # First determine if this item contains a nested list.
        # If it does, remove the list from the item's children (so the item content can be
        # properly rendered) and process it into editor_block_data if it does
        for i, child in enumerate(marko_list_item_block.children):
            if isinstance(child, List):
                nested_marko_list_block = marko_list_item_block.children.pop(i)
                for nested_child in nested_marko_list_block.children:
                    nested_list_item_editor_block_data_kwargs, _ = \
                        cls._marko_list_item_block_to_editor_block_data_dict(nested_child)
                    list_item_editor_block_data_kwargs['items'].append(nested_list_item_editor_block_data_kwargs)

                break

        # Marko wraps the text content in li when it renders a ListItem into html,
        # so we need to remove that, plus remove any leftover tags (e.g. if the content
        # in the li was a header)
        content = cls._remove_outer_html_tags(
            cls._remove_outer_html_tags(marko.render(marko_list_item_block), 'li')
        )

        is_checklist_item, is_checked, content = \
            cls._try_parse_content_as_checklist_item(content)

        if is_checklist_item:
            # Each editor list item block data tracks its checked state through an
            # item level meta
            list_item_editor_block_data_kwargs['meta']['checked'] = is_checked

        list_item_editor_block_data_kwargs['content'] = content

        list_editor_block_data_style = None

        if parent_marko_list_block:
            # We determine the entire list style from the top level list items for simplicity.
            # If parent_marko_list_block is present, it means we are processing the top level list.
            list_editor_block_data_style = ListEditorBlockDataStyle.UNORDERED
            if is_checklist_item:
                # The presence of checklist items takes priority over an ordered list style
                list_editor_block_data_style = ListEditorBlockDataStyle.CHECKLIST
            elif parent_marko_list_block.ordered:
                list_editor_block_data_style = ListEditorBlockDataStyle.ORDERED

        return list_item_editor_block_data_kwargs, list_editor_block_data_style

    @classmethod
    def _marko_block_to_editor_block_data_dict(cls, marko_block: BlockElement | Element):
        if isinstance(marko_block, BlankLine):
            return { 'text': '' }

        if isinstance(marko_block, Paragraph):
            editor_block_text_string = cls._remove_outer_html_tags(marko.render(marko_block))
            return { 'text': editor_block_text_string }

        if isinstance(marko_block, Heading):
            return {
                'text': cls._remove_outer_html_tags(marko.render(marko_block)),
                'level': marko_block.level,
            }

        if isinstance(marko_block, List):
            list_editor_block_data_dict = {
                'items': [],
                'style': ListEditorBlockDataStyle.UNORDERED,
                'meta': {},
            }

            for child in marko_block.children:
                list_item_editor_block_data_dict, list_editor_block_data_style = \
                    cls._marko_list_item_block_to_editor_block_data_dict(child, marko_block)

                list_editor_block_data_dict['items'].append(list_item_editor_block_data_dict)
                list_editor_block_data_dict['style'] = list_editor_block_data_style

            return list_editor_block_data_dict

        logging.warning(
            f'Unsupported marko block type: {marko_block.__class__.__name__!r}. '
            f'Rendering content to html and adding to markdown as a basic paragraph block.'
        )

        return { 'text': marko.render(marko_block) }


    @classmethod
    def marko_block_to_markdown_string(cls, marko_block: BlockElement) -> str:
        return cls.html_to_markdown(marko.render(marko_block))



from __future__ import annotations

from django.urls import path

from django_spire.knowledge.entry.version.views import json_views


app_name = 'json'

urlpatterns = [
    path(
        '<int:pk>/update_blocks/',
        json_views.update_blocks_view,
        name='update_blocks',
    ),
    path(
        '<int:pk>/update_entry_from_version/',
        json_views.update_entry_from_version_view,
        name='update_entry_from_version',
    ),
]



from __future__ import annotations

from django.contrib.contenttypes.models import ContentType

from django_spire.comment.models import Comment
from django_spire.core.tests.test_cases import BaseTestCase


class TestCommentModel(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.content_type = ContentType.objects.get_for_model(self.super_user)
        self.comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='This is a test comment.'
        )

    def test_str(self) -> None:
        assert str(self.comment) == 'This is a ...'

    def test_str_short_information(self) -> None:
        comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Short'
        )
        assert str(comment) == 'Short...'

    def test_scrape_username_list(self) -> None:
        comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Hello @john_doe and @jane_doe'
        )
        usernames = comment.scrape_username_list()
        assert usernames == ['john_doe', 'jane_doe']

    def test_scrape_username_list_empty(self) -> None:
        usernames = self.comment.scrape_username_list()
        assert usernames == []

    def test_scrape_username_list_single(self) -> None:
        comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Hello @stratus'
        )
        usernames = comment.scrape_username_list()
        assert usernames == ['stratus']

    def test_find_user_list(self) -> None:
        comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Hello @stratus'
        )
        users = comment.find_user_list()
        assert self.super_user in users

    def test_find_user_list_empty(self) -> None:
        users = self.comment.find_user_list()
        assert users.count() == 0

    def test_update(self) -> None:
        self.comment.update('Updated information')
        self.comment.refresh_from_db()
        assert self.comment.information == 'Updated information'
        assert self.comment.is_edited is True

    def test_update_sets_is_edited(self) -> None:
        assert self.comment.is_edited is False
        self.comment.update('New content')
        assert self.comment.is_edited is True

    def test_parent_relationship(self) -> None:
        parent_comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Parent comment'
        )
        child_comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            parent=parent_comment,
            information='Child comment'
        )
        assert child_comment.parent == parent_comment
        assert child_comment in parent_comment.children.all()

    def test_default_ordering(self) -> None:
        comment1 = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='First'
        )

        comment2 = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Second'
        )

        comments = list(Comment.objects.order_by('-pk')[:2])
        assert comments == [comment2, comment1]


class TestCommentRelationships(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.content_type = ContentType.objects.get_for_model(self.super_user)
        self.comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Test comment'
        )

    def test_comment_user_relationship(self) -> None:
        assert self.comment.user == self.super_user
        assert self.comment in self.super_user.comment_list.all()

    def test_comment_content_object(self) -> None:
        assert self.comment.content_object == self.super_user
        assert self.comment.object_id == self.super_user.pk

    def test_nested_replies(self) -> None:
        parent_comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Parent'
        )
        child1 = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            parent=parent_comment,
            information='Child 1'
        )
        child2 = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            parent=parent_comment,
            information='Child 2'
        )
        assert parent_comment.children.count() == 2
        assert child1 in parent_comment.children.all()
        assert child2 in parent_comment.children.all()

    def test_cascade_delete_parent(self) -> None:
        parent_comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Parent'
        )
        child_comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            parent=parent_comment,
            information='Child'
        )
        child_pk = child_comment.pk
        parent_comment.delete()
        assert Comment.objects.filter(pk=child_pk).exists() is False


class TestCommentDefaults(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.content_type = ContentType.objects.get_for_model(self.super_user)
        self.comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Test comment'
        )

    def test_default_is_active(self) -> None:
        assert self.comment.is_active is True

    def test_default_is_deleted(self) -> None:
        assert self.comment.is_deleted is False

    def test_default_is_edited(self) -> None:
        assert self.comment.is_edited is False

    def test_default_information(self) -> None:
        comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user
        )
        assert comment.information == ''

    def test_created_datetime_auto_set(self) -> None:
        assert self.comment.created_datetime is not None

    def test_parent_default_null(self) -> None:
        assert self.comment.parent is None



from __future__ import annotations

from django.db import models

from django_spire.core.tag.models import Tag
from django_spire.core.tag.tools import simplify_and_weight_tag_set_to_dict, simplify_tag_set


class TagModelMixin(models.Model):
    tags = models.ManyToManyField(Tag, related_name='+', blank=True, editable=False)

    class Meta:
        abstract = True

    @property
    def tag_set(self) -> set[str]:
        return set(self.tags.values_list('name', flat=True))

    @property
    def simplified_tag_set(self) -> set[str]:
        return simplify_tag_set(self.tag_set)

    @property
    def simplified_and_weighted_tag_dict(self) -> dict[str, int]:
        return simplify_and_weight_tag_set_to_dict(self.tag_set)



from __future__ import annotations

from django_spire.auth.group.models import AuthGroup


def bulk_create_groups_from_names(names: list[str]) -> list[AuthGroup]:
    existing_groups = set(AuthGroup.objects.all().values_list('name', flat=True))
    seen_names = set()

    new_groups = []

    for name in names:
        if name in existing_groups or name in seen_names:
            continue

        sanitized_name = name.replace('\x00', '')

        if sanitized_name in existing_groups or sanitized_name in seen_names:
            continue

        seen_names.add(sanitized_name)
        new_groups.append(AuthGroup(name=sanitized_name))

    return AuthGroup.objects.bulk_create(new_groups)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.exceptions import PermissionDenied
from django.test import RequestFactory

import pytest

from django_spire.auth.controller.controller import AppAuthController, BaseAuthController
from django_spire.auth.controller.exceptions import (
    AuthControllerNotFoundError,
    AuthControllerRequestError,
)
from django_spire.auth.user.tests.factories import create_super_user, create_user
from django_spire.core.tests.test_cases import BaseTestCase

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


class AuthControllerForTest(BaseAuthController):
    condition: bool = True

    def also_can_access(self) -> bool:
        return True

    def can_access(self) -> bool:
        return True

    def cannot_access(self) -> bool:
        return False

    def conditional_access(self) -> bool:
        return self.condition

    def request_based_access(self) -> bool:
        return self.request.user.is_authenticated


class BaseAuthControllerTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.factory = RequestFactory()
        self.user = create_user(username='testuser')
        self.controller = BaseAuthController()

    def test_init_without_request(self) -> None:
        controller = BaseAuthController()
        assert controller._request is None

    def test_init_with_request(self) -> None:
        request = self.factory.get('/test/')
        controller = BaseAuthController(request=request)
        assert controller._request == request

    def test_request_property_raises_when_none(self) -> None:
        controller = BaseAuthController()
        with pytest.raises(AuthControllerRequestError) as exc_info:
            _ = controller.request
        assert 'None' in str(exc_info.value)

    def test_request_setter(self) -> None:
        controller = BaseAuthController()
        request = self.factory.get('/test/')
        controller.request = request
        assert controller.request == request

    def test_request_setter_multiple_times(self) -> None:
        controller = BaseAuthController()
        request1 = self.factory.get('/test1/')
        request2 = self.factory.get('/test2/')
        controller.request = request1
        controller.request = request2
        assert controller.request == request2

    def test_request_property_returns_same_request(self) -> None:
        request = self.factory.get('/test/')
        controller = BaseAuthController(request=request)
        assert controller.request is request

    def test_request_setter_with_none_then_valid(self) -> None:
        controller = BaseAuthController()
        request = self.factory.get('/test/')
        controller._request = None
        controller.request = request
        assert controller.request == request


class BaseAuthControllerPermissionRequiredTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.factory = RequestFactory()
        self.user = create_user(username='testuser')
        self.controller = AuthControllerForTest()

    def test_permission_required_with_callable_permission_passes(self) -> None:
        @self.controller.permission_required('can_access')
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = self.user
        result = test_view(request)
        assert result == 'success'

    def test_permission_required_callable_fails(self) -> None:
        @self.controller.permission_required('cannot_access')
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = self.user

        with pytest.raises(PermissionDenied):
            test_view(request)

    def test_permission_required_any_with_callable(self) -> None:
        @self.controller.permission_required('can_access', 'cannot_access', all_required=False)
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = self.user
        result = test_view(request)
        assert result == 'success'

    def test_permission_required_all_callable_one_fails(self) -> None:
        @self.controller.permission_required('can_access', 'cannot_access', all_required=True)
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = self.user

        with pytest.raises(PermissionDenied):
            test_view(request)

    def test_permission_required_sets_request(self) -> None:
        @self.controller.permission_required('can_access')
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = self.user
        test_view(request)
        assert self.controller.request == request

    def test_permission_required_preserves_function_name(self) -> None:
        @self.controller.permission_required('can_access')
        def my_special_view(_request: WSGIRequest) -> str:
            return 'success'

        assert my_special_view.__name__ == 'my_special_view'

    def test_permission_required_with_args(self) -> None:
        @self.controller.permission_required('can_access')
        def test_view(_request: WSGIRequest, arg1: str, arg2: str) -> str:
            return f'{arg1}-{arg2}'

        request = self.factory.get('/test/')
        request.user = self.user
        result = test_view(request, 'hello', 'world')
        assert result == 'hello-world'

    def test_permission_required_with_kwargs(self) -> None:
        @self.controller.permission_required('can_access')
        def test_view(_request: WSGIRequest, **kwargs) -> str:
            return kwargs.get('key', 'default')

        request = self.factory.get('/test/')
        request.user = self.user
        result = test_view(request, key='value')
        assert result == 'value'

    def test_multiple_callable_permissions_all_pass(self) -> None:
        @self.controller.permission_required('can_access', 'also_can_access', all_required=True)
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = self.user
        result = test_view(request)
        assert result == 'success'

    def test_conditional_permission(self) -> None:
        @self.controller.permission_required('conditional_access')
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = self.user

        self.controller.condition = True
        result = test_view(request)
        assert result == 'success'

    def test_conditional_permission_fails(self) -> None:
        self.controller.condition = False

        @self.controller.permission_required('conditional_access')
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = self.user

        with pytest.raises(PermissionDenied):
            test_view(request)

    def test_permission_required_with_mixed_args_kwargs(self) -> None:
        @self.controller.permission_required('can_access')
        def test_view(_request: WSGIRequest, arg1: str, kwarg1: str | None = None) -> str:
            return f'{arg1}-{kwarg1}'

        request = self.factory.get('/test/')
        request.user = self.user
        result = test_view(request, 'positional', kwarg1='keyword')
        assert result == 'positional-keyword'

    def test_permission_required_callable_uses_request(self) -> None:
        @self.controller.permission_required('request_based_access')
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = self.user
        result = test_view(request)
        assert result == 'success'

    def test_permission_required_with_string_permission_fallback(self) -> None:
        @self.controller.permission_required('nonexistent_method', all_required=True)
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = self.user

        with pytest.raises(PermissionDenied):
            test_view(request)

    def test_permission_required_superuser_string_permission(self) -> None:
        superuser = create_super_user()

        @self.controller.permission_required('some.fake_permission')
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = superuser
        result = test_view(request)
        assert result == 'success'

    def test_permission_required_empty_permissions(self) -> None:
        @self.controller.permission_required()
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = self.user
        result = test_view(request)
        assert result == 'success'

    def test_permission_required_any_all_fail(self) -> None:
        @self.controller.permission_required('cannot_access', all_required=False)
        def test_view(_request: WSGIRequest) -> str:
            return 'success'

        request = self.factory.get('/test/')
        request.user = self.user

        with pytest.raises(PermissionDenied):
            test_view(request)

    def test_permission_required_view_returns_none(self) -> None:
        @self.controller.permission_required('can_access')
        def test_view(_request: WSGIRequest) -> None:
            return None

        request = self.factory.get('/test/')
        request.user = self.user
        result = test_view(request)
        assert result is None

    def test_permission_required_view_returns_complex_object(self) -> None:
        @self.controller.permission_required('can_access')
        def test_view(_request: WSGIRequest) -> dict:
            return {'key': 'value', 'list': [1, 2, 3]}

        request = self.factory.get('/test/')
        request.user = self.user
        result = test_view(request)
        assert result == {'key': 'value', 'list': [1, 2, 3]}


class AppAuthControllerTestCase(BaseTestCase):
    def test_app_controller_not_found_raises_error(self) -> None:
        with pytest.raises(AuthControllerNotFoundError):
            AppAuthController('nonexistent_app')

    def test_app_controller_not_found_error_message(self) -> None:
        with pytest.raises(AuthControllerNotFoundError) as exc_info:
            AppAuthController('nonexistent_app')

        assert 'nonexistent_app' in str(exc_info.value)


class AuthControllerExceptionsTestCase(BaseTestCase):
    def test_auth_controller_request_error_is_exception(self) -> None:
        assert issubclass(AuthControllerRequestError, Exception)

    def test_auth_controller_not_found_error_is_exception(self) -> None:
        assert issubclass(AuthControllerNotFoundError, Exception)

    def test_auth_controller_request_error_message(self) -> None:
        error = AuthControllerRequestError('Custom message')
        assert str(error) == 'Custom message'

    def test_auth_controller_not_found_error_message(self) -> None:
        error = AuthControllerNotFoundError('App not found')
        assert str(error) == 'App not found'



import json

from django_spire.contrib.choices.choices import SpireTextChoices
from django_spire.core.tests.test_cases import BaseTestCase


class TestSpireTextChoices(BaseTestCase):
    def setUp(self):
        class StatusChoices(SpireTextChoices):
            DRAFT = ('dra', 'Draft')
            PUBLISHED = ('pub', 'Published')
            ARCHIVED = ('arc', 'Archived')


        self.StatusChoices = StatusChoices

    def test_inherits_from_text_choices(self):
        assert issubclass(self.StatusChoices, SpireTextChoices), self.StatusChoices.__class__

    def test_choices_property_exists(self):
        assert hasattr(self.StatusChoices, 'choices'), self.StatusChoices.dir()
        assert self.StatusChoices.choices is not None, self.StatusChoices.choices

    def test_to_glue_choices_returns_string(self):
        result = self.StatusChoices.to_glue_choices()
        assert isinstance(result, str), type(result)

    def test_to_glue_choices_returns_valid_json(self):
        result = self.StatusChoices.to_glue_choices()
        try:
            parsed = json.loads(result)
        except json.JSONDecodeError:
            assert 'to_glue_choices did not return valid JSON'

    def test_to_glue_choices_correct_structure(self):
        result = self.StatusChoices.to_glue_choices()
        parsed = json.loads(result)

        assert isinstance(parsed, list), type(parsed)
        assert len(parsed) == 3

        assert parsed[0] == ['dra', 'Draft']
        assert parsed[1] == ['pub', 'Published']
        assert parsed[2] == ['arc', 'Archived']

    def test_empty_choices(self):
        class EmptyChoices(SpireTextChoices):
            pass


        result = EmptyChoices.to_glue_choices()
        parsed = json.loads(result)
        assert parsed == []

    def test_single_choice(self):
        class SingleChoice(SpireTextChoices):
            ONLY = 'only', 'Only Option'


        result = SingleChoice.to_glue_choices()
        parsed = json.loads(result)
        assert parsed == [['only', 'Only Option']]



from enum import StrEnum


class ColumnType(StrEnum):
    TEXT = 'text'
    CHOICE = 'choice'
    NUMBER = 'number'
    NUMBER_1 = 'decimal_1'
    NUMBER_2 = 'decimal_2'
    NUMBER_3 = 'decimal_3'
    DOLLAR = 'dollar'
    DOLLAR_1 = 'dollar_1'
    DOLLAR_2 = 'dollar_2'
    DOLLAR_3 = 'dollar_3'
    PERCENT = 'percent'
    PERCENT_1 = 'percent_1'
    PERCENT_2 = 'percent_2'
    PERCENT_3 = 'percent_3'





from __future__ import annotations

from django.test import TestCase

from django_spire.core.tag.models import Tag
from django_spire.knowledge.collection.models import Collection


class TestTags(TestCase):
    def setUp(self) -> None:
        self.collection = Collection.objects.create(
            name='Test Collection',
            description='Test Collection Description'
        )

        self.tag_set_a = {'apple', 'banana', 'orange'}
        self.tag_set_b = {'grape', 'kiwi', 'pear'}
        self.tag_set_c = {'lemon', 'lime', 'mango'}

        self.mixed_tag_set = {'apple', 'grape', 'lemon'}

        self.dirty_tag_set = {'HaT~', 'four* tacos', '123 hoop!!'}

        self.collection.services.tag.add_tags_from_tag_set(self.tag_set_a)
        self.collection.services.tag.add_tags_from_tag_set(self.tag_set_b)
        self.collection.services.tag.add_tags_from_tag_set(self.tag_set_c)

        self.collection.services.tag.add_tags_from_tag_set(self.mixed_tag_set)

        self.collection.services.tag.add_tags_from_tag_set(self.dirty_tag_set)

    def test_tag_add_and_set(self) -> None:
        self.collection.services.tag.set_tags_from_tag_set(self.mixed_tag_set)

        assert 'lemon' in self.collection.tag_set
        assert 'banana' not in self.collection.tag_set

    def test_tag_bool_methods(self) -> None:
        self.collection.services.tag.remove_tags_by_tag_set(self.tag_set_b)

        assert self.collection.services.tag.has_tags_in_tag_set(self.tag_set_c) is True
        assert self.collection.services.tag.has_tags_in_tag_set(self.mixed_tag_set) is False

    def test_tag_count(self) -> None:
        assert Tag.objects.count() == 12
        assert self.collection.tags.count() == 12

    def test_tag_matching_methods(self) -> None:
        assert len(self.collection.services.tag.get_matching_tags_from_tag_set(self.tag_set_a)) == 3

        self.collection.services.tag.remove_tags_by_tag_set(self.tag_set_a)

        assert len(self.collection.services.tag.get_matching_tags_from_tag_set(self.tag_set_a)) == 0
        assert len(self.collection.services.tag.get_matching_tags_from_tag_set(self.tag_set_b)) == 3
        assert len(self.collection.services.tag.get_matching_tags_from_tag_set(self.mixed_tag_set)) == 2

        self.collection.services.tag.remove_tags_by_tag_set(self.tag_set_b)

        assert len(self.collection.services.tag.get_matching_tags_from_tag_set(self.tag_set_b)) == 0
        assert len(self.collection.services.tag.get_matching_tags_from_tag_set(self.mixed_tag_set)) == 1

    def test_tag_metric_methods(self) -> None:
        assert abs(
            self.collection.services.tag.get_matching_percentage_of_tag_set(self.tag_set_a) - 1.00
        ) < 0.01

        assert abs(
            self.collection.services.tag.get_matching_percentage_of_model_tags_from_tag_set(self.tag_set_b) - 0.25
        ) < 0.01

        self.collection.services.tag.remove_tags_by_tag_set(self.mixed_tag_set)

        assert abs(
            self.collection.services.tag.get_matching_percentage_of_tag_set(self.tag_set_a) - 0.666
        ) < 0.01

        assert abs(
            self.collection.services.tag.get_matching_percentage_of_model_tags_from_tag_set(self.tag_set_b) - 0.222
        ) < 0.01

    def test_tag_removal(self) -> None:
        assert 'apple' in self.collection.tag_set

        self.collection.services.tag.remove_tags_by_tag_set(self.mixed_tag_set)

        assert 'lemon' not in self.collection.tag_set
        assert 'kiwi' in self.collection.tag_set

    def test_tag_removal_count(self) -> None:
        self.collection.services.tag.remove_tags_by_tag_set(self.tag_set_a)

        assert self.collection.tags.count() == 9



from __future__ import annotations

from django.contrib.contenttypes.models import ContentType

from django_spire.comment.models import Comment
from django_spire.core.tests.test_cases import BaseTestCase


class TestCommentQuerySet(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.content_type = ContentType.objects.get_for_model(self.super_user)
        self.comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Test comment'
        )

    def test_active(self) -> None:
        comments = Comment.objects.active()
        assert self.comment in comments

    def test_active_excludes_deleted(self) -> None:
        comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Deleted comment',
            is_deleted=True
        )
        comments = Comment.objects.active()
        assert comment not in comments

    def test_top_level(self) -> None:
        parent_comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Parent'
        )
        child_comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            parent=parent_comment,
            information='Child'
        )
        top_level = Comment.objects.top_level()
        assert parent_comment in top_level
        assert child_comment not in top_level

    def test_prefetch_user(self) -> None:
        comments = Comment.objects.prefetch_user()
        assert self.comment in comments

    def test_prefetch_parent(self) -> None:
        parent_comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Parent'
        )
        child_comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            parent=parent_comment,
            information='Child'
        )
        comments = Comment.objects.prefetch_parent()
        assert child_comment in comments

    def test_prefetch_replies(self) -> None:
        parent_comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Parent'
        )
        comments = Comment.objects.prefetch_replies()
        assert parent_comment in comments

    def test_chained_querysets(self) -> None:
        parent_comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            information='Parent'
        )
        child_comment = Comment.objects.create(
            content_type=self.content_type,
            object_id=self.super_user.pk,
            user=self.super_user,
            parent=parent_comment,
            information='Child'
        )
        comments = Comment.objects.active().top_level().prefetch_user()
        assert parent_comment in comments
        assert child_comment not in comments



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from django_spire.file.models import File
    from django_spire.knowledge.entry.version.models import EntryVersion
    from django_spire.knowledge.entry.version.block.models import EntryVersionBlock


class BaseConverter(ABC):
    def __init__(self, entry_version: EntryVersion):
        self.entry_version = entry_version

    @abstractmethod
    def convert_file_to_blocks(self, file: File) -> list[EntryVersionBlock]:
        raise NotImplementedError



from __future__ import annotations

from unittest.mock import MagicMock

from django.contrib.admin.sites import AdminSite
from django.test import TestCase

from django_spire.history.activity.admin import ActivityAdmin
from django_spire.history.activity.models import Activity
from django_spire.history.admin import HistoryEventAdmin
from django_spire.history.models import HistoryEvent
from django_spire.history.viewed.admin import ViewAdmin
from django_spire.history.viewed.models import Viewed


class TestActivityAdmin(TestCase):
    def setUp(self) -> None:
        self.site = AdminSite()
        self.admin = ActivityAdmin(Activity, self.site)

    def test_information_snippet_long(self) -> None:
        activity = MagicMock()
        activity.information = 'This is a very long information string that exceeds twenty characters'

        result = self.admin.information_snippet(activity)

        assert result.endswith('...')
        assert len(result) == 23

    def test_information_snippet_none(self) -> None:
        activity = MagicMock()
        activity.information = None

        result = self.admin.information_snippet(activity)

        assert result == 'No Information'

    def test_information_snippet_short(self) -> None:
        activity = MagicMock()
        activity.information = 'Short info'

        result = self.admin.information_snippet(activity)

        assert result == 'Short info'

    def test_list_display(self) -> None:
        assert 'id' in self.admin.list_display
        assert 'content_object_link' in self.admin.list_display
        assert 'verb' in self.admin.list_display
        assert 'user_link' in self.admin.list_display

    def test_list_filter(self) -> None:
        assert 'verb' in self.admin.list_filter
        assert 'created_datetime' in self.admin.list_filter

    def test_recipient_link_no_recipient(self) -> None:
        activity = MagicMock()
        activity.recipient = None

        result = self.admin.recipient_link(activity)

        assert result == 'No Recipient'


class TestHistoryEventAdmin(TestCase):
    def setUp(self) -> None:
        self.site = AdminSite()
        self.admin = HistoryEventAdmin(HistoryEvent, self.site)

    def test_list_display(self) -> None:
        assert 'id' in self.admin.list_display
        assert 'content_object_link' in self.admin.list_display
        assert 'content_type' in self.admin.list_display
        assert 'event_verbose' in self.admin.list_display

    def test_list_filter(self) -> None:
        assert 'event' in self.admin.list_filter
        assert 'created_datetime' in self.admin.list_filter


class TestViewedAdmin(TestCase):
    def setUp(self) -> None:
        self.site = AdminSite()
        self.admin = ViewAdmin(Viewed, self.site)

    def test_list_display(self) -> None:
        assert 'id' in self.admin.list_display
        assert 'content_object_link' in self.admin.list_display
        assert 'user_link' in self.admin.list_display
        assert 'created_datetime' in self.admin.list_display

    def test_list_filter(self) -> None:
        assert 'created_datetime' in self.admin.list_filter



from __future__ import annotations

from django_spire.knowledge.collection.models import Collection
from django_spire.knowledge.entry import models

from django_spire.contrib.seeding import DjangoModelSeeder
from django_spire.knowledge.entry.version.seeding.seeder import EntryVersionSeeder


class EntrySeeder(DjangoModelSeeder):
    model_class = models.Entry
    cache_name = 'entry_seeder'
    fields = {
        'id': 'exclude',
        'current_version': 'exclude',
        'collection_id': ('custom', 'fk_random', {'model_class': Collection}),
        'name': (
            'llm',
            'A name for a document. Make it fun and give it a theme'
        ),
    }

    @classmethod
    def _correct_order(cls, entries: list[models.Entry]) -> list[models.Entry]:
        for collection in Collection.objects.all():
            collection_entries = collection.entries.all()

            for idx, entry in enumerate(collection_entries):
                entry.order = idx

            cls.model_class.objects.bulk_update(collection_entries, ['order'])

        return entries

    @classmethod
    def seed_database(
        cls,
        count: int = 1,
        fields: dict | None = None
    ) -> list[models.Entry]:
        entries = super().seed_database(
            count=count,
            fields=fields
        )

        cls._correct_order(entries)
        entries = cls._set_current_version(entries=entries, count=count)

        for entry in entries:
            entry.services.tag.process_and_set_tags()

        return entries

    @classmethod
    def _set_current_version(cls, entries: list[models.Entry], count: int = 1):
        entry_versions = EntryVersionSeeder.seed_database(count=count)

        for entry, entry_version in zip(entries, entry_versions, strict=False):
            entry_version.entry = entry
            entry.current_version = entry_version

        cls.model_class.objects.bulk_update(entries, ['current_version'])
        EntryVersionSeeder.model_class.objects.bulk_update(entry_versions, ['entry'])
        return entries



from __future__ import annotations

from django.apps import AppConfig

from django_spire.utils import check_required_apps


class AiConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    label = 'django_spire_ai'
    name = 'django_spire.ai'

    REQUIRED_APPS = ('django_spire_core',)
    URLPATTERNS_INCLUDE = 'django_spire.ai.urls'
    URLPATTERNS_NAMESPACE = 'ai'

    def ready(self) -> None:
        check_required_apps(self.label)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.db.models import Q

from django_spire.contrib.ordering.querysets import OrderingQuerySetMixin
from django_spire.history.querysets import HistoryQuerySet
from django_spire.knowledge.entry.version.choices import EntryVersionStatusChoices

if TYPE_CHECKING:
    from django.db.models import QuerySet

    from django_spire.auth.user.models import AuthUser
    from django_spire.knowledge.entry.models import Entry


class EntryQuerySet(HistoryQuerySet, OrderingQuerySetMixin):
    def get_by_version_block_id(self, version_block_id: int) -> Entry:
        return self.get(
            current_version__block__id=version_block_id
        )

    def has_current_version(self) -> QuerySet[Entry]:
        return self.filter(current_version__isnull=False)

    def id_in(self, ids: list[int]) -> QuerySet[Entry]:
        return self.filter(id__in=ids)

    def user_has_access(self, user: AuthUser) -> QuerySet[Entry]:
        return self.filter(
            Q(
                current_version__status=EntryVersionStatusChoices.PUBLISHED
            ) |
            Q(
                current_version__author=user,
                current_version__status=EntryVersionStatusChoices.DRAFT
            )
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib.auth.decorators import login_required
from django.urls import reverse

from django_spire.contrib.generic_views import portal_views

from django_spire.notification.app.models import AppNotification

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from django.template.response import TemplateResponse


@login_required()
def app_notification_list_view(request: WSGIRequest) -> TemplateResponse:
    return portal_views.list_view(
        request,
        context_data={'notification_endpoint': reverse('django_spire:notification:app:template:scroll_items')},
        model=AppNotification,
        template='django_spire/notification/app/page/list_page.html'
    )



from __future__ import annotations

from django.contrib.auth.models import Group

from django_spire.auth.user.tests.factories import create_user
from django_spire.auth.user.tools import add_user_to_all_user_group
from django_spire.core.tests.test_cases import BaseTestCase


class AddUserToAllUserGroupTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.user = create_user(username='testuser')

    def test_creates_all_users_group_if_not_exists(self) -> None:
        Group.objects.filter(name='All Users').delete()
        add_user_to_all_user_group(self.user)
        assert Group.objects.filter(name='All Users').exists()

    def test_adds_user_to_existing_group(self) -> None:
        group = Group.objects.create(name='All Users')
        add_user_to_all_user_group(self.user)
        assert self.user in group.user_set.all()

    def test_user_in_all_users_group(self) -> None:
        add_user_to_all_user_group(self.user)
        group = Group.objects.get(name='All Users')
        assert self.user in group.user_set.all()

    def test_idempotent_add(self) -> None:
        add_user_to_all_user_group(self.user)
        add_user_to_all_user_group(self.user)
        group = Group.objects.get(name='All Users')
        assert group.user_set.filter(pk=self.user.pk).count() == 1

    def test_multiple_users_added(self) -> None:
        user2 = create_user(username='testuser2')
        add_user_to_all_user_group(self.user)
        add_user_to_all_user_group(user2)
        group = Group.objects.get(name='All Users')
        assert group.user_set.count() >= 2
        assert self.user in group.user_set.all()
        assert user2 in group.user_set.all()

    def test_user_remains_in_other_groups(self) -> None:
        other_group = Group.objects.create(name='Other Group')
        self.user.groups.add(other_group)
        add_user_to_all_user_group(self.user)
        assert other_group in self.user.groups.all()

    def test_group_created_with_exact_name(self) -> None:
        Group.objects.filter(name='All Users').delete()
        add_user_to_all_user_group(self.user)
        group = Group.objects.get(name='All Users')
        assert group.name == 'All Users'

    def test_inactive_user_can_be_added(self) -> None:
        inactive_user = create_user(username='inactive', is_active=False)
        add_user_to_all_user_group(inactive_user)
        group = Group.objects.get(name='All Users')
        assert inactive_user in group.user_set.all()

    def test_does_not_affect_other_groups(self) -> None:
        other_group = Group.objects.create(name='Other Group')
        initial_count = other_group.user_set.count()
        add_user_to_all_user_group(self.user)
        assert other_group.user_set.count() == initial_count

    def test_user_not_added_to_other_groups(self) -> None:
        other_group = Group.objects.create(name='Other Group')
        add_user_to_all_user_group(self.user)
        assert self.user not in other_group.user_set.all()

    def test_multiple_calls_single_group_entry(self) -> None:
        for _ in range(5):
            add_user_to_all_user_group(self.user)
        group = Group.objects.get(name='All Users')
        assert group.user_set.filter(pk=self.user.pk).count() == 1

    def test_group_persists_after_function_call(self) -> None:
        Group.objects.filter(name='All Users').delete()
        add_user_to_all_user_group(self.user)
        assert Group.objects.filter(name='All Users').exists()

    def test_user_membership_persists(self) -> None:
        add_user_to_all_user_group(self.user)
        self.user.refresh_from_db()
        group = Group.objects.get(name='All Users')
        assert group in self.user.groups.all()

    def test_does_not_remove_existing_members(self) -> None:
        existing_user = create_user(username='existinguser')
        group = Group.objects.create(name='All Users')
        group.user_set.add(existing_user)

        add_user_to_all_user_group(self.user)

        assert existing_user in group.user_set.all()
        assert self.user in group.user_set.all()

    def test_works_with_staff_user(self) -> None:
        staff_user = create_user(username='staffuser', is_staff=True)
        add_user_to_all_user_group(staff_user)
        group = Group.objects.get(name='All Users')
        assert staff_user in group.user_set.all()

    def test_works_with_superuser(self) -> None:
        superuser = create_user(username='superuser', is_superuser=True)
        add_user_to_all_user_group(superuser)
        group = Group.objects.get(name='All Users')
        assert superuser in group.user_set.all()

    def test_case_sensitive_group_name(self) -> None:
        Group.objects.create(name='all users')
        Group.objects.filter(name='All Users').delete()
        add_user_to_all_user_group(self.user)
        assert Group.objects.filter(name='All Users').exists()
        assert Group.objects.filter(name='all users').exists()

    def test_many_users_added(self) -> None:
        users = [create_user(username=f'manyuser{i}') for i in range(20)]
        for user in users:
            add_user_to_all_user_group(user)
        group = Group.objects.get(name='All Users')
        for user in users:
            assert user in group.user_set.all()

    def test_user_groups_query(self) -> None:
        add_user_to_all_user_group(self.user)
        groups = self.user.groups.all()
        group_names = [g.name for g in groups]
        assert 'All Users' in group_names

    def test_group_user_count_increments(self) -> None:
        Group.objects.filter(name='All Users').delete()
        group = Group.objects.create(name='All Users')
        initial_count = group.user_set.count()

        add_user_to_all_user_group(self.user)
        group.refresh_from_db()

        assert group.user_set.count() == initial_count + 1

    def test_does_not_create_duplicate_groups(self) -> None:
        Group.objects.filter(name='All Users').delete()
        user1 = create_user(username='user1')
        user2 = create_user(username='user2')

        add_user_to_all_user_group(user1)
        add_user_to_all_user_group(user2)

        assert Group.objects.filter(name='All Users').count() == 1



from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest

from django_spire.auth.user.tests.factories import create_user
from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.choices import (
    NotificationStatusChoices,
    NotificationTypeChoices,
)
from django_spire.notification.exceptions import NotificationError
from django_spire.notification.models import Notification
from django_spire.notification.sms.processor import SMSNotificationProcessor
from django_spire.notification.sms.tests.factories import create_test_sms_notification


class SMSNotificationProcessorTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.user = create_user(username='test_sms_processor_user')
        self.processor = SMSNotificationProcessor()

    @patch('django_spire.notification.sms.processor.Client')
    @patch('django_spire.notification.sms.processor.TwilioSMSHelper')
    def test_process_sets_status_to_sent(
        self,
        mock_helper_class: MagicMock,
        mock_client_class: MagicMock
    ):
        mock_helper = MagicMock()
        mock_helper_class.return_value = mock_helper

        sms_notification = create_test_sms_notification(user=self.user)
        notification = sms_notification.notification

        self.processor.process(notification)

        notification.refresh_from_db()
        assert notification.status == NotificationStatusChoices.SENT

    @patch('django_spire.notification.sms.processor.Client')
    @patch('django_spire.notification.sms.processor.TwilioSMSHelper')
    def test_process_sets_sent_datetime(
        self,
        mock_helper_class: MagicMock,
        mock_client_class: MagicMock
    ):
        mock_helper = MagicMock()
        mock_helper_class.return_value = mock_helper

        sms_notification = create_test_sms_notification(user=self.user)
        notification = sms_notification.notification

        self.processor.process(notification)

        notification.refresh_from_db()
        assert notification.sent_datetime is not None

    @patch('django_spire.notification.sms.processor.Client')
    @patch('django_spire.notification.sms.processor.TwilioSMSHelper')
    def test_process_calls_send(
        self,
        mock_helper_class: MagicMock,
        mock_client_class: MagicMock
    ):
        mock_helper = MagicMock()
        mock_helper_class.return_value = mock_helper

        sms_notification = create_test_sms_notification(user=self.user)
        notification = sms_notification.notification

        self.processor.process(notification)

        mock_helper.send.assert_called_once()

    def test_process_raises_error_for_wrong_type(self):
        notification = Notification.objects.create(
            user=self.user,
            type=NotificationTypeChoices.APP,
            title='Test',
            body='Test',
            status=NotificationStatusChoices.PENDING,
        )

        with pytest.raises(NotificationError):
            self.processor.process(notification)

    @patch('django_spire.notification.sms.processor.Client')
    @patch('django_spire.notification.sms.processor.BulkTwilioSMSHelper')
    def test_process_list(
        self,
        mock_helper_class: MagicMock,
        mock_client_class: MagicMock
    ):
        mock_helper = MagicMock()
        mock_helper_class.return_value = mock_helper

        notifications = [
            create_test_sms_notification(user=self.user).notification
            for _ in range(3)
        ]

        self.processor.process_list(notifications)

        mock_helper.send_notifications.assert_called_once()



from __future__ import annotations

import pytest

from django_spire.ai.chat.message_intel import BaseMessageIntel, DefaultMessageIntel
from django_spire.core.tests.test_cases import BaseTestCase


class CustomMessageIntel(BaseMessageIntel):
    _template: str = 'django_spire/ai/chat/message/default_message.html'
    extra_data: str = 'Extra'
    text: str

    def render_to_str(self) -> str:
        return f'{self.text} - {self.extra_data}'


class TestMessageIntel(BaseTestCase):
    def test_default_message_intel_has_template(self) -> None:
        intel = DefaultMessageIntel(text='Test')

        assert intel.template == 'django_spire/ai/chat/message/default_message.html'

    def test_default_message_intel_render_to_str(self) -> None:
        intel = DefaultMessageIntel(text='Hello World')
        result = intel.render_to_str()

        assert result == 'Hello World'

    def test_custom_message_intel_render_to_str(self) -> None:
        intel = CustomMessageIntel(text='Test', extra_data='Custom')
        result = intel.render_to_str()

        assert result == 'Test - Custom'

    def test_render_template_to_str_renders_django_template(self) -> None:
        intel = DefaultMessageIntel(text='Hello')
        result = intel.render_template_to_str()

        assert isinstance(result, str)
        assert len(result) > 0

    def test_render_template_to_str_with_context(self) -> None:
        intel = DefaultMessageIntel(text='Test')
        result = intel.render_template_to_str(context_data={'extra': 'data'})

        assert isinstance(result, str)

    def test_template_property(self) -> None:
        intel = DefaultMessageIntel(text='Test')

        assert intel.template == intel._template

    def test_message_intel_raises_without_template(self) -> None:
        with pytest.raises(ValueError):
            class NoTemplateIntel(BaseMessageIntel):
                _template = None

                def render_to_str(self) -> str:
                    return 'test'

    def test_message_intel_raises_with_empty_template(self) -> None:
        with pytest.raises(ValueError):
            class EmptyTemplateIntel(BaseMessageIntel):
                _template = ''

                def render_to_str(self) -> str:
                    return 'test'

    def test_default_message_intel_model_dump(self) -> None:
        intel = DefaultMessageIntel(text='Test text')
        dump = intel.model_dump()

        assert 'text' in dump
        assert dump['text'] == 'Test text'

    def test_custom_message_intel_default_extra_data(self) -> None:
        intel = CustomMessageIntel(text='Test')

        assert intel.extra_data == 'Extra'

    def test_default_message_intel_empty_text(self) -> None:
        intel = DefaultMessageIntel(text='')
        result = intel.render_to_str()

        assert result == ''

    def test_default_message_intel_long_text(self) -> None:
        long_text = 'A' * 10000
        intel = DefaultMessageIntel(text=long_text)
        result = intel.render_to_str()

        assert result == long_text
        assert len(result) == 10000

    def test_default_message_intel_special_characters(self) -> None:
        special_text = '<script>alert("xss")</script>'
        intel = DefaultMessageIntel(text=special_text)
        result = intel.render_to_str()

        assert result == special_text

    def test_default_message_intel_unicode(self) -> None:
        unicode_text = 'Hello 世界 🌍'
        intel = DefaultMessageIntel(text=unicode_text)
        result = intel.render_to_str()

        assert result == unicode_text

    def test_render_template_to_str_with_none_context(self) -> None:
        intel = DefaultMessageIntel(text='Test')
        result = intel.render_template_to_str(context_data=None)

        assert isinstance(result, str)

    def test_render_template_to_str_with_empty_context(self) -> None:
        intel = DefaultMessageIntel(text='Test')
        result = intel.render_template_to_str(context_data={})

        assert isinstance(result, str)

    def test_custom_message_intel_override_extra_data(self) -> None:
        intel = CustomMessageIntel(text='Test', extra_data='Override')

        assert intel.extra_data == 'Override'
        assert intel.render_to_str() == 'Test - Override'



from __future__ import annotations

import inspect

from typing import Any, Callable, Sequence


def get_object_from_module_string(module_string: str) -> Any:
    try:
        module_string, object_name = module_string.rsplit('.', 1)
        module = __import__(module_string, fromlist=[object_name])
    except ImportError as e:
        message = f'Could not import module: {module_string}'
        raise ImportError(message) from e

    return getattr(module, object_name)

def get_callable_from_module_string_and_validate_arguments(
    module_string: str,
    valid_args: Sequence[str]
) -> Callable:
    callable_ = get_object_from_module_string(module_string)

    if not callable(callable_):
        message = f'Object {module_string} is not callable'
        raise TypeError(message)

    sig = inspect.signature(callable_)
    params = sig.parameters

    for name in valid_args:
        if name not in params:
            message = f'{callable_.__qualname__} is missing required argument: {name}'
            raise TypeError(message)

    return callable_



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.file.utils import random_64_char_token


class Random64CharTokenTests(BaseTestCase):
    def test_returns_string(self):
        result = random_64_char_token()

        assert isinstance(result, str)

    def test_returns_64_characters(self):
        result = random_64_char_token()

        assert len(result) == 64

    def test_returns_hexadecimal(self):
        result = random_64_char_token()

        int(result, 16)

    def test_returns_unique_values(self):
        result1 = random_64_char_token()
        result2 = random_64_char_token()

        assert result1 != result2

    def test_returns_lowercase(self):
        result = random_64_char_token()

        assert result == result.lower()



from __future__ import annotations

from django_spire.auth.controller.controller import BaseAuthController


class BaseReportAuthController(BaseAuthController):
    def can_add(self):
        return self.request.user.has_perm('django_spire_metric_report.add_reportrun')

    def can_change(self):
        return self.request.user.has_perm('django_spire_metric_report.change_reportrun')

    def can_delete(self):
        return self.request.user.has_perm('django_spire_metric_report.delete_reportrun')

    def can_view(self):
        return self.request.user.has_perm('django_spire_metric_report.view_reportrun')



from __future__ import annotations

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.timezone import localtime

from django_spire.history.choices import HistoryEventChoices


class HistoryEvent(models.Model):
    content_type = models.ForeignKey(
        ContentType,
        related_name='history_events',
        related_query_name='history_event',
        on_delete=models.CASCADE,
        editable=False
    )

    object_id = models.PositiveIntegerField(editable=False)
    content_object = GenericForeignKey('content_type', 'object_id')

    event = models.CharField(max_length=4, choices=HistoryEventChoices.choices)

    created_datetime = models.DateTimeField(default=localtime)

    def __str__(self) -> str:
        return f'{self.content_object} - {self.event_verbose}'

    @property
    def event_verbose(self) -> str:
        return dict(HistoryEventChoices.choices)[self.event]

    class Meta:
        db_table = 'django_spire_history_event'
        verbose_name = 'History Event'
        verbose_name_plural = 'History Events'




from __future__ import annotations

from django.db import models

from django.contrib import admin


class AiUsageMixin(models.Model):
    event_count = models.IntegerField(default=0, verbose_name='Events')
    token_usage = models.IntegerField(default=0, verbose_name='Tokens')
    run_time_seconds = models.FloatField(default=0.0, verbose_name='Run Time')
    was_successful = models.BooleanField(default=True, verbose_name='Success')

    class Meta:
        abstract = True


class AiUsageAdminMixin(admin.ModelAdmin):
    def run_time_seconds_formatted(self, obj) -> str:
        return f"{obj.run_time_seconds:.3f}s"

    run_time_seconds_formatted.short_description = 'Run Time'



from __future__ import annotations

from django.core.management.base import BaseCommand

from django_spire.knowledge.entry.models import Entry


class Command(BaseCommand):
    help = 'Rebuilds the search index for all knowledge base entries.'

    def handle(self, *args, **options):
        self.stdout.write('Rebuilding search indexes...')

        Entry.services.search_index.rebuild_all_search_indexes()

        self.stdout.write(self.style.SUCCESS('Search indexes rebuilt successfully.'))



from __future__ import annotations

from django.urls import path

from django_spire.auth.group.views import form_views


app_name = 'form'

# Group
urlpatterns = [
    path('group/form', form_views.form_view, name='add'),
    path('group/<int:pk>/form', form_views.form_view, name='update'),
    path('group/<int:pk>/delete/form', form_views.delete_form_view, name='delete'),
]

# Group User
urlpatterns += [
    path('group/<int:pk>/user/form', form_views.user_form_view, name='user'),
    path('group/<int:group_pk>/user/<int:pk>/form',
         form_views.group_remove_user_form_view,
         name='user_remove'),
]



from __future__ import annotations

import json
import string

from unittest.mock import MagicMock, patch

from django.template import Context, RequestContext, Template
from django.test import RequestFactory, TestCase

from django_spire.core.templatetags.json import to_json
from django_spire.core.templatetags.message import django_messages_to_json
from django_spire.core.templatetags.spire_core_tags import (
    add_str,
    content_type_url,
    generate_id,
    in_list,
    index,
    is_path,
    not_in_list,
    query_param_url,
    safe_dict_items,
    to_snake_case,
)
from django_spire.core.templatetags.string_formating import (
    dashes_and_spaces_to_underscore,
    dashes_to_underscore,
    spaces_to_underscore,
    underscores_to_spaces,
)
from django_spire.core.templatetags.variable_types import (
    is_dict,
    is_list,
    is_list_or_tuple,
    is_not_dict,
    is_not_list,
    is_not_list_or_tuple,
    is_not_tuple,
    is_tuple,
)


class TestAddStr(TestCase):
    def test_concatenates_strings(self) -> None:
        assert add_str('Hello', 'World') == 'HelloWorld'

    def test_empty_strings(self) -> None:
        assert add_str('', '') == ''

    def test_with_spaces(self) -> None:
        assert add_str('Hello ', 'World') == 'Hello World'


class TestContentTypeUrl(TestCase):
    def test_constructs_url_with_metadata(self) -> None:
        class Dummy:
            pass

        dummy = Dummy()
        dummy._meta = type(
            'meta',
            (),
            {'app_label': 'myapp', 'model_name': 'dummy'}
        )

        func = 'django_spire.core.templatetags.spire_core_tags.reverse'
        return_value = 'http://example.com/dummy'

        with patch(func, return_value=return_value) as mock_reverse:
            url = content_type_url('dummy_url', dummy)

            mock_reverse.assert_called_once_with(
                'dummy_url',
                kwargs={'app_label': 'myapp', 'model_name': 'dummy'}
            )

            assert url == 'http://example.com/dummy'


class TestDashesAndSpacesToUnderscore(TestCase):
    def test_dashes_and_spaces(self) -> None:
        assert dashes_and_spaces_to_underscore('hello-world test') == 'hello_world_test'

    def test_no_changes_needed(self) -> None:
        assert dashes_and_spaces_to_underscore('hello_world') == 'hello_world'

    def test_only_dashes(self) -> None:
        assert dashes_and_spaces_to_underscore('hello-world') == 'hello_world'

    def test_only_spaces(self) -> None:
        assert dashes_and_spaces_to_underscore('hello world') == 'hello_world'


class TestDashesToUnderscore(TestCase):
    def test_multiple_dashes(self) -> None:
        assert dashes_to_underscore('hello-world-test') == 'hello_world_test'

    def test_no_dashes(self) -> None:
        assert dashes_to_underscore('hello') == 'hello'

    def test_single_dash(self) -> None:
        assert dashes_to_underscore('hello-world') == 'hello_world'


class TestDjangoMessagesToJson(TestCase):
    def test_empty_messages(self) -> None:
        result = django_messages_to_json([])
        assert json.loads(result) == []

    def test_multiple_messages(self) -> None:
        messages = [
            MagicMock(message='First', level_tag='info'),
            MagicMock(message='Second', level_tag='error'),
        ]

        result = django_messages_to_json(messages)
        parsed = json.loads(result)

        assert len(parsed) == 2
        assert parsed[0]['message'] == 'First'
        assert parsed[0]['type'] == 'info'
        assert parsed[1]['message'] == 'Second'
        assert parsed[1]['type'] == 'error'

    def test_single_message(self) -> None:
        messages = [MagicMock(message='Test message', level_tag='success')]

        result = django_messages_to_json(messages)
        parsed = json.loads(result)

        assert len(parsed) == 1
        assert parsed[0]['message'] == 'Test message'
        assert parsed[0]['type'] == 'success'
        assert len(parsed[0]['id']) == 8


class TestGenerateId(TestCase):
    def test_generates_8_characters(self) -> None:
        identifier = generate_id()
        assert len(identifier) == 8

    def test_only_ascii_letters(self) -> None:
        identifier = generate_id()
        assert all(ch in string.ascii_letters for ch in identifier)

    def test_unique_ids(self) -> None:
        ids = {generate_id() for _ in range(100)}
        assert len(ids) == 100


class TestInList(TestCase):
    def test_empty_list(self) -> None:
        assert in_list('a', '') is False

    def test_value_in_list(self) -> None:
        assert in_list('a', 'a,b,c') is True

    def test_value_not_in_list(self) -> None:
        assert in_list('d', 'a,b,c') is False


class TestIndex(TestCase):
    def test_index_out_of_bounds(self) -> None:
        items = [10, 20, 30]
        assert index(items, 5) == items

    def test_negative_index(self) -> None:
        items = [10, 20, 30]
        assert index(items, -1) == 30

    def test_valid_index(self) -> None:
        items = [10, 20, 30]
        assert index(items, 1) == 20


class TestIsDict(TestCase):
    def test_dict(self) -> None:
        assert is_dict({'key': 'value'}) is True

    def test_empty_dict(self) -> None:
        assert is_dict({}) is True

    def test_list(self) -> None:
        assert is_dict([]) is False

    def test_string(self) -> None:
        assert is_dict('string') is False


class TestIsList(TestCase):
    def test_dict(self) -> None:
        assert is_list({}) is False

    def test_empty_list(self) -> None:
        assert is_list([]) is True

    def test_list(self) -> None:
        assert is_list([1, 2, 3]) is True

    def test_tuple(self) -> None:
        assert is_list((1, 2, 3)) is False


class TestIsListOrTuple(TestCase):
    def test_dict(self) -> None:
        assert is_list_or_tuple({}) is False

    def test_list(self) -> None:
        assert is_list_or_tuple([1, 2, 3]) is True

    def test_string(self) -> None:
        assert is_list_or_tuple('string') is False

    def test_tuple(self) -> None:
        assert is_list_or_tuple((1, 2, 3)) is True


class TestIsNotDict(TestCase):
    def test_dict(self) -> None:
        assert is_not_dict({'key': 'value'}) is False

    def test_list(self) -> None:
        assert is_not_dict([]) is True

    def test_string(self) -> None:
        assert is_not_dict('string') is True


class TestIsNotList(TestCase):
    def test_dict(self) -> None:
        assert is_not_list({}) is True

    def test_list(self) -> None:
        assert is_not_list([1, 2, 3]) is False

    def test_tuple(self) -> None:
        assert is_not_list((1, 2, 3)) is True


class TestIsNotListOrTuple(TestCase):
    def test_dict(self) -> None:
        assert is_not_list_or_tuple({}) is True

    def test_list(self) -> None:
        assert is_not_list_or_tuple([1, 2, 3]) is False

    def test_string(self) -> None:
        assert is_not_list_or_tuple('string') is True

    def test_tuple(self) -> None:
        assert is_not_list_or_tuple((1, 2, 3)) is False


class TestIsNotTuple(TestCase):
    def test_list(self) -> None:
        assert is_not_tuple([1, 2, 3]) is True

    def test_tuple(self) -> None:
        assert is_not_tuple((1, 2, 3)) is False


class TestIsPath(TestCase):
    def test_empty_current(self) -> None:
        assert is_path('', '/test') is False

    def test_empty_url(self) -> None:
        assert is_path('/test', '') is False

    def test_exact_match(self) -> None:
        assert is_path('/test', '/test') is True

    def test_root_url(self) -> None:
        assert is_path('/test', '/') is False

    def test_starts_with(self) -> None:
        assert is_path('/test/page', '/test') is True


class TestIsTuple(TestCase):
    def test_list(self) -> None:
        assert is_tuple([1, 2, 3]) is False

    def test_tuple(self) -> None:
        assert is_tuple((1, 2, 3)) is True


class TestNotInList(TestCase):
    def test_empty_list(self) -> None:
        assert not_in_list('a', '') is True

    def test_value_in_list(self) -> None:
        assert not_in_list('a', 'a,b,c') is False

    def test_value_not_in_list(self) -> None:
        assert not_in_list('x', 'a,b,c') is True


class TestQueryParamUrl(TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.factory = RequestFactory()

    def test_multiple_params(self) -> None:
        request = self.factory.get('/some_path', {'foo': 'bar', 'baz': 'qux'})
        context = RequestContext(request, {})

        func = 'django_spire.core.templatetags.spire_core_tags.reverse'
        return_value = 'http://example.com/dummy'

        with patch(func, return_value=return_value) as mock_reverse:
            url = query_param_url(context, 'dummy_url')
            mock_reverse.assert_called_once_with('dummy_url', kwargs={})

            assert url.startswith('http://example.com/dummy?')
            assert 'foo=bar' in url
            assert 'baz=qux' in url

    def test_no_params(self) -> None:
        request = self.factory.get('/some_path')
        context = RequestContext(request, {})

        func = 'django_spire.core.templatetags.spire_core_tags.reverse'
        return_value = 'http://example.com/dummy'

        with patch(func, return_value=return_value):
            url = query_param_url(context, 'dummy_url')

            assert url == 'http://example.com/dummy?'


class TestSafeDictItems(TestCase):
    def test_dict_with_items_key(self) -> None:
        d = {'items': 'value', 'other': 'data'}
        result = list(safe_dict_items(d))

        assert ('items', 'value') in result
        assert ('other', 'data') in result

    def test_no_items_attribute(self) -> None:
        result = safe_dict_items('not a dict')
        assert result == []

    def test_normal_dict(self) -> None:
        d = {'key': 'value'}
        result = list(safe_dict_items(d))

        assert result == [('key', 'value')]


class TestSpacesToUnderscore(TestCase):
    def test_multiple_spaces(self) -> None:
        assert spaces_to_underscore('hello world test') == 'hello_world_test'

    def test_no_spaces(self) -> None:
        assert spaces_to_underscore('hello') == 'hello'

    def test_single_space(self) -> None:
        assert spaces_to_underscore('hello world') == 'hello_world'


class TestToJson(TestCase):
    def test_dict(self) -> None:
        result = to_json({'key': 'value'})
        assert result == '{"key": "value"}'

    def test_invalid_value(self) -> None:
        result = to_json(object())
        assert result == ''

    def test_list(self) -> None:
        result = to_json([1, 2, 3])
        assert result == '[1, 2, 3]'

    def test_nested_structure(self) -> None:
        result = to_json({'list': [1, 2, 3], 'nested': {'key': 'value'}})
        parsed = json.loads(result)

        assert parsed['list'] == [1, 2, 3]
        assert parsed['nested'] == {'key': 'value'}


class TestToSnakeCase(TestCase):
    def test_already_snake_case(self) -> None:
        assert to_snake_case('hello_world') == 'hello_world'

    def test_multiple_spaces(self) -> None:
        assert to_snake_case('Hello World Test') == 'hello_world_test'

    def test_simple_conversion(self) -> None:
        assert to_snake_case('Hello World') == 'hello_world'


class TestUnderscoresToSpaces(TestCase):
    def test_multiple_underscores(self) -> None:
        assert underscores_to_spaces('hello_world_test') == 'hello world test'

    def test_no_underscores(self) -> None:
        assert underscores_to_spaces('hello') == 'hello'

    def test_single_underscore(self) -> None:
        assert underscores_to_spaces('hello_world') == 'hello world'


class TemplateRenderingTests(TestCase):
    def test_render_add_str_filter(self) -> None:
        template_code = """
            {% load spire_core_tags %}

            {{ "Hello" | add_str:" World" }}
        """

        tmpl = Template(template_code)
        rendered = tmpl.render(Context())
        assert 'Hello World' in rendered

    def test_render_to_snake_case_tag(self) -> None:
        template_code = """
            {% load spire_core_tags %}

            {% to_snake_case "Hello World" as snake_case %}
            {{ snake_case }}
        """

        tmpl = Template(template_code)
        rendered = tmpl.render(Context())
        assert 'hello_world' in rendered



from __future__ import annotations

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.test import RequestFactory

from django_spire.auth.user.tests.factories import create_user
from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.knowledge.auth.controller import BaseKnowledgeAuthController
from django_spire.knowledge.collection.models import Collection


class BaseKnowledgeAuthControllerTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()
        self.user = create_user(username='test_auth_user')
        self.request = self.factory.get('/')
        self.request.user = self.user

        content_type = ContentType.objects.get_for_model(Collection)

        self.view_permission = Permission.objects.get(
            codename='view_collection',
            content_type=content_type
        )
        self.add_permission = Permission.objects.get(
            codename='add_collection',
            content_type=content_type
        )
        self.change_permission = Permission.objects.get(
            codename='change_collection',
            content_type=content_type
        )
        self.delete_permission = Permission.objects.get(
            codename='delete_collection',
            content_type=content_type
        )
        self.access_all_permission = Permission.objects.get(
            codename='can_access_all_collections',
            content_type=content_type
        )
        self.change_groups_permission = Permission.objects.get(
            codename='can_change_collection_groups',
            content_type=content_type
        )

    def _refresh_user_and_request(self):
        self.user = type(self.user).objects.get(pk=self.user.pk)
        self.request.user = self.user

    def test_can_view_without_permission(self):
        controller = BaseKnowledgeAuthController(self.request)
        assert controller.can_view() is False

    def test_can_view_with_permission(self):
        self.user.user_permissions.add(self.view_permission)
        self._refresh_user_and_request()

        controller = BaseKnowledgeAuthController(self.request)
        assert controller.can_view() is True

    def test_can_add_without_permission(self):
        controller = BaseKnowledgeAuthController(self.request)
        assert controller.can_add() is False

    def test_can_add_with_permission(self):
        self.user.user_permissions.add(self.add_permission)
        self._refresh_user_and_request()

        controller = BaseKnowledgeAuthController(self.request)
        assert controller.can_add() is True

    def test_can_change_without_permission(self):
        controller = BaseKnowledgeAuthController(self.request)
        assert controller.can_change() is False

    def test_can_change_with_permission(self):
        self.user.user_permissions.add(self.change_permission)
        self._refresh_user_and_request()

        controller = BaseKnowledgeAuthController(self.request)
        assert controller.can_change() is True

    def test_can_delete_without_permission(self):
        controller = BaseKnowledgeAuthController(self.request)
        assert controller.can_delete() is False

    def test_can_delete_with_permission(self):
        self.user.user_permissions.add(self.delete_permission)
        self._refresh_user_and_request()

        controller = BaseKnowledgeAuthController(self.request)
        assert controller.can_delete() is True

    def test_can_access_all_collections_without_permission(self):
        controller = BaseKnowledgeAuthController(self.request)
        assert controller.can_access_all_collections() is False

    def test_can_access_all_collections_with_permission(self):
        self.user.user_permissions.add(self.access_all_permission)
        self._refresh_user_and_request()

        controller = BaseKnowledgeAuthController(self.request)
        assert controller.can_access_all_collections() is True

    def test_can_change_collection_groups_without_permission(self):
        controller = BaseKnowledgeAuthController(self.request)
        assert controller.can_change_collection_groups() is False

    def test_can_change_collection_groups_with_permission(self):
        self.user.user_permissions.add(self.change_groups_permission)
        self._refresh_user_and_request()

        controller = BaseKnowledgeAuthController(self.request)
        assert controller.can_change_collection_groups() is True



from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.contrib.service import BaseDjangoModelService

if TYPE_CHECKING:
    from django_spire.knowledge.collection.models import Collection


class CollectionOrderingService(BaseDjangoModelService['Collection']):
    obj: Collection

    def reorder(self, order: int, new_parent_pk: int | None):
        current_parent = self.obj.parent
        if current_parent is None:
            origin_objects = self.obj_class.objects.parentless().active()

        else:
            origin_objects = (
                self.obj_class.objects
                .by_parent(parent=current_parent)
                .active()
            )

        if new_parent_pk is None:
            destination_objects = self.obj_class.objects.parentless().active()

        else:
            new_parent = self.obj_class.objects.get(pk=new_parent_pk)
            destination_objects = new_parent.children.active()

        self.obj.ordering_services.processor.move_to_position(
            destination_objects=destination_objects,
            position=order,
            origin_objects=origin_objects
        )

        self.obj, _ = self.obj.services.save_model_obj(parent_id=new_parent_pk)



from __future__ import annotations

import json
from typing import TYPE_CHECKING

from django.contrib.auth.models import AnonymousUser
from django.urls import reverse

from django_spire.contrib.generic_views.portal_views import infinite_scrolling_view
from django.template.response import TemplateResponse
from django_spire.notification.app.models import AppNotification


if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest

def notification_infinite_scroll_view(request: WSGIRequest) -> TemplateResponse:
    if isinstance(request.user, AnonymousUser):
        notifications = []

    else:
        notifications = (
            AppNotification.objects.active()
            .is_sent()
            .annotate_is_viewed_by_user(request.user)
            .select_related('notification')
            .distinct()
            .ordered_by_priority_and_sent_datetime()
        )

    body_data = json.loads(request.body.decode('utf-8'))

    return infinite_scrolling_view(
        request,
        queryset=notifications,
        queryset_name='notifications',
        context_data={
            'app_notification_list_url': body_data.get('app_notification_list_url'),
        },
        template='django_spire/notification/app/scroll/item/items.html',
    )


def notification_dropdown_template_view(request: WSGIRequest) -> TemplateResponse:
    body_data = json.loads(request.body.decode('utf-8'))

    return TemplateResponse(
        request,
        context={
            'app_notification_list_url': body_data.get('app_notification_list_url'),
            'notification_endpoint': reverse('django_spire:notification:app:template:scroll_items')
        },
        template='django_spire/notification/app/dropdown/notification_dropdown_content.html'
    )



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.knowledge.entry.version.block.data.list.choices import ListEditorBlockDataStyle
from django_spire.knowledge.entry.version.block.data.list.maps import (
    LIST_BLOCK_DATA_META_MAP,
    LIST_BLOCK_DATA_REVERSE_META_MAP,
)
from django_spire.knowledge.entry.version.block.data.list.meta import (
    ChecklistItemMeta,
    OrderedListItemMeta,
)


class ListBlockDataMapTests(BaseTestCase):
    def test_meta_map_ordered(self):
        assert LIST_BLOCK_DATA_META_MAP[ListEditorBlockDataStyle.ORDERED] == OrderedListItemMeta

    def test_meta_map_checklist(self):
        assert LIST_BLOCK_DATA_META_MAP[ListEditorBlockDataStyle.CHECKLIST] == ChecklistItemMeta

    def test_meta_map_unordered(self):
        assert LIST_BLOCK_DATA_META_MAP[ListEditorBlockDataStyle.UNORDERED] is None

    def test_reverse_meta_map_ordered(self):
        assert LIST_BLOCK_DATA_REVERSE_META_MAP[OrderedListItemMeta] == ListEditorBlockDataStyle.ORDERED

    def test_reverse_meta_map_checklist(self):
        assert LIST_BLOCK_DATA_REVERSE_META_MAP[ChecklistItemMeta] == ListEditorBlockDataStyle.CHECKLIST

    def test_reverse_meta_map_unordered(self):
        assert LIST_BLOCK_DATA_REVERSE_META_MAP[None] == ListEditorBlockDataStyle.UNORDERED



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.push.models import PushNotification


class PushNotificationModelTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.push_notification = PushNotification.objects.create()

    def test_created_datetime_auto_set(self):
        assert self.push_notification.created_datetime is not None

    def test_is_active_default_true(self):
        assert self.push_notification.is_active is True

    def test_is_deleted_default_false(self):
        assert self.push_notification.is_deleted is False

    def test_meta_verbose_name(self):
        assert PushNotification._meta.verbose_name == 'Push Notification'

    def test_meta_verbose_name_plural(self):
        assert PushNotification._meta.verbose_name_plural == 'Push Notifications'

    def test_meta_db_table(self):
        assert PushNotification._meta.db_table == 'django_spire_notification_push'



from __future__ import annotations

import hashlib

from django_spire.conf import settings


def get_theme_cookie_name() -> str:
    if not hasattr(settings, 'BASE_DIR'):
        return 'django-spire-theme'

    encoded = str(settings.BASE_DIR).encode()
    identifier = hashlib.md5(encoded, usedforsecurity=False).hexdigest()[:8]
    return f'django-spire-theme-{identifier}'



from __future__ import annotations

import json

from abc import abstractmethod
from typing import TYPE_CHECKING

from django.db.models import QuerySet, Q
from django_spire.contrib.form.utils import show_form_errors
from django_spire.contrib.queryset.enums import SessionFilterActionEnum
from django_spire.contrib.session.controller import SessionController

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from django.forms import Form


class SessionFilterQuerySetMixin(QuerySet):
    def process_session_filter(
        self,
        request: WSGIRequest,
        session_key: str,
        form_class: type[Form],
        is_from_body: bool = False
    ) -> QuerySet:
        # Session keys must match to process new queryset data

        data = json.loads(request.body.decode('utf-8')) if is_from_body else request.GET

        try:
            action = SessionFilterActionEnum(data.get('action'))
        except ValueError:
            action = None

        form = form_class(data)

        if form.is_valid():
            session = SessionController(request=request, session_key=session_key)

            if action == SessionFilterActionEnum.CLEAR:
                session.purge()
                return self

            # Apply filters when the user submits the filter form
            if (
                action == SessionFilterActionEnum.FILTER
                and session_key == data.get('session_filter_key')
            ):
                # Update session data
                for key, value in form.cleaned_data.items():
                    session.add_data(key, value)

            # If the session is expired, return the unfiltered queryset
            if session.is_expired:
                return self

            # When no new filter data is applied and session is NOT yet expired,
            # return the original queryset
            return self.bulk_filter(session.data)

        show_form_errors(request, form)
        return self

    @abstractmethod
    def bulk_filter(self, filter_data: dict) -> QuerySet:
        pass


class SearchQuerySetMixin(QuerySet):
    def search(self, value: str | None) -> QuerySet:
        words = value.split(' ')

        filtered_query = self

        char_fields = [
            field.name for field in self.model._meta.fields
            if field.get_internal_type() == 'CharField'
        ]

        for word in words:
            or_conditions = Q()
            for field in char_fields:
                or_conditions |= Q(**{f"{field}__icontains": word})
            filtered_query = filtered_query.filter(or_conditions)

        return filtered_query


class ChoicesQueryMixin(QuerySet):
    def to_choices(self):
        return [
            (obj.pk, str(obj)) for obj in self
        ]



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.knowledge.entry.version.block.choices import BlockTypeChoices
from django_spire.knowledge.entry.version.block.data.list.choices import (
    ListEditorBlockDataStyle,
    OrderedListCounterType,
)


class BlockTypeChoicesTests(BaseTestCase):
    def test_text_value(self):
        assert BlockTypeChoices.TEXT == 'text'

    def test_heading_value(self):
        assert BlockTypeChoices.HEADING == 'heading'

    def test_list_value(self):
        assert BlockTypeChoices.LIST == 'list'

    def test_choices_count(self):
        assert len(BlockTypeChoices.choices) == 3


class ListEditorBlockDataStyleTests(BaseTestCase):
    def test_unordered_value(self):
        assert ListEditorBlockDataStyle.UNORDERED == 'unordered'

    def test_ordered_value(self):
        assert ListEditorBlockDataStyle.ORDERED == 'ordered'

    def test_checklist_value(self):
        assert ListEditorBlockDataStyle.CHECKLIST == 'checklist'

    def test_choices_count(self):
        assert len(ListEditorBlockDataStyle.choices) == 3


class OrderedListCounterTypeTests(BaseTestCase):
    def test_numeric_value(self):
        assert OrderedListCounterType.NUMERIC == 'numeric'

    def test_upper_roman_value(self):
        assert OrderedListCounterType.UPPER_ROMAN == 'upper-roman'

    def test_lower_roman_value(self):
        assert OrderedListCounterType.LOWER_ROMAN == 'lower-roman'

    def test_upper_alpha_value(self):
        assert OrderedListCounterType.UPPER_ALPHA == 'upper-alpha'

    def test_lower_alpha_value(self):
        assert OrderedListCounterType.LOWER_ALPHA == 'lower-alpha'

    def test_choices_count(self):
        assert len(OrderedListCounterType.choices) == 5



from __future__ import annotations

from django.contrib.auth.models import User

from django_spire.help_desk.choices import (
    HelpDeskTicketPriorityChoices,
    HelpDeskTicketPurposeChoices,
    HelpDeskTicketStatusChoices,
)
from django_spire.help_desk.models import HelpDeskTicket


def create_test_helpdesk_ticket(**kwargs) -> HelpDeskTicket:
    if 'created_by' not in kwargs:
        kwargs['created_by'] = User.objects.first()

    return HelpDeskTicket.objects.create(**create_test_helpdesk_ticket_data(**kwargs))


def create_test_helpdesk_ticket_data(**kwargs) -> dict:
    ticket_data = {
        'priority': HelpDeskTicketPriorityChoices.LOW,
        'purpose': HelpDeskTicketPurposeChoices.APP,
        'status': HelpDeskTicketStatusChoices.READY,
        'description': 'This is a test ticket.',
    }

    ticket_data.update(**kwargs)

    return ticket_data



from __future__ import annotations

from django_spire.knowledge.entry.version.block.data.data import BaseEditorJsBlockData


class HeadingEditorBlockData(BaseEditorJsBlockData):
    text: str
    level: int

    def render_to_text(self) -> str:
        from django_spire.knowledge.entry.version.converters.markdown_converter import \
            MarkdownConverter

        text = MarkdownConverter.html_to_markdown(self.text)
        return f'{"#" * self.level} {text}\n'



from __future__ import annotations

from django.test import TestCase

from django_spire.contrib.seeding.field.enums import FieldSeederTypesEnum
from django_spire.contrib.seeding.model.enums import ModelSeederDefaultsEnum


class TestFieldSeederTypesEnum(TestCase):
    def test_callable_value(self) -> None:
        assert FieldSeederTypesEnum.CALLABLE == 'callable'

    def test_custom_value(self) -> None:
        assert FieldSeederTypesEnum.CUSTOM == 'custom'

    def test_faker_value(self) -> None:
        assert FieldSeederTypesEnum.FAKER == 'faker'

    def test_is_str_enum(self) -> None:
        assert isinstance(FieldSeederTypesEnum.LLM, str)

    def test_llm_value(self) -> None:
        assert FieldSeederTypesEnum.LLM == 'llm'

    def test_static_value(self) -> None:
        assert FieldSeederTypesEnum.STATIC == 'static'


class TestModelSeederDefaultsEnum(TestCase):
    def test_faker_value(self) -> None:
        assert ModelSeederDefaultsEnum.FAKER == 'faker'

    def test_included_value(self) -> None:
        assert ModelSeederDefaultsEnum.INCLUDED == 'included'

    def test_is_str_enum(self) -> None:
        assert isinstance(ModelSeederDefaultsEnum.LLM, str)

    def test_llm_value(self) -> None:
        assert ModelSeederDefaultsEnum.LLM == 'llm'



from __future__ import annotations

from random import shuffle
from typing import TYPE_CHECKING

from dandy import Prompt

if TYPE_CHECKING:
    from django.db.models import Model


def hierarchical_selection_prompt(
    model_class: type[Model],
    self_reference_field: str,
    constraint: list[str]
) -> Prompt:
    # You must create/call a seeding function/prompt that
    # generates a "parent" or a "parent" must exist before
    # this prompt can be used.

    parent_instances = list(model_class.objects.all())
    shuffle(parent_instances)

    parent_constraints = []

    for instance in parent_instances:
        constraint_values = []

        for field in constraint:
            value = getattr(instance, field, None)
            constraint_values.append(f"{field}: {value!s}")

        parent_constraints.append(f'ID {instance.id}: {", ".join(constraint_values)}')

    model_name = model_class._meta.verbose_name.title()
    model_plural = model_class._meta.verbose_name_plural.title()

    constraint_descriptions = [f'`{field}`' for field in constraint]
    constraint_text = ' and '.join(constraint_descriptions)

    return (
        Prompt()
        .text(f'When creating {model_plural} with a `{self_reference_field}`, you must:')
        .list([
            f'Select one of the existing {model_plural} as the `{self_reference_field}`',
            f'Each child must have the same {constraint_text} as its `{self_reference_field}`',
        ])
        .line_break()

        .text(f'The available {model_plural} with their constraint values:')
        .list(parent_constraints)
        .line_break()

        .text(f'Do NOT create a parent. For example, no `{model_name}` should have `parent=null`.')
        .text(f'All rows should have a(n) {constraint_text}')
        .line_break()

        .text('Children MUST inherit these values from their parent. Do not modify them.')
    )



from django import template
from django.db.models import Model

register = template.Library()


@register.filter
def model_app_label(model_obj: Model) -> str:
    """
    Return the Django app label for the given model instance.

    Args:
        model_obj: A Django model instance.

    Returns:
        The app label defined by the model's app configuration.
    """

    return model_obj._meta.app_label


@register.filter
def model_name(model_obj: Model) -> str:
    """
    Return the Django model name for the given model instance.

    Args:
        model_obj: A Django model instance.

    Returns:
        The model name (lowercased) as defined by Django's model metadata.
    """

    return model_obj._meta.model_name



from __future__ import annotations

from django.test import TestCase

from django_spire.consts import (
    __VERSION__,
    MAINTENANCE_MODE_SETTINGS_NAME,
    NOTIFICATION_THROTTLE_RATE_PER_MINUTE_SETTINGS_NAME,
)


class TestConsts(TestCase):
    def test_maintenance_mode_settings_name(self) -> None:
        assert MAINTENANCE_MODE_SETTINGS_NAME == 'MAINTENANCE_MODE'

    def test_notification_throttle_rate_settings_name(self) -> None:
        assert NOTIFICATION_THROTTLE_RATE_PER_MINUTE_SETTINGS_NAME == 'NOTIFICATION_THROTTLE_RATE_PER_MINUTE'

    def test_version_is_string(self) -> None:
        assert isinstance(__VERSION__, str)

    def test_version_format(self) -> None:
        parts = __VERSION__.split('.')

        assert len(parts) == 3

        for part in parts:
            assert part.isdigit()



from __future__ import annotations

from django_spire.contrib.service import BaseDjangoModelService

from typing import TYPE_CHECKING

from django_spire.knowledge.collection.services.factory_service import CollectionGroupFactoryService
from django_spire.knowledge.collection.services.ordering_service import CollectionOrderingService
from django_spire.knowledge.collection.services.processor_service import CollectionProcessorService
from django_spire.knowledge.collection.services.tag_service import CollectionTagService
from django_spire.knowledge.collection.services.tool_services import CollectionToolService
from django_spire.knowledge.collection.services.transformation_service import CollectionTransformationService

if TYPE_CHECKING:
    from django_spire.knowledge.collection.models import Collection, CollectionGroup


class CollectionService(BaseDjangoModelService['Collection']):
    obj: Collection

    ordering = CollectionOrderingService()
    processor = CollectionProcessorService()
    tag = CollectionTagService()
    tool = CollectionToolService()
    transformation = CollectionTransformationService()

    def save_model_obj(self, **field_data) -> tuple[Collection, bool]:
        self.obj, created = super().save_model_obj(**field_data)

        if self.obj.parent_id is None:
            destination_objects = self.obj_class.objects.parentless().active()
        else:
            destination_objects = (
                self.obj_class.objects
                .by_parent_id(parent_id=self.obj.parent_id)
                .active()
            )

        self.obj.ordering_services.processor.move_to_position(
            destination_objects=destination_objects,
            position=0 if created else self.obj.order,
        )

        return self.obj, created


class CollectionGroupService(BaseDjangoModelService['CollectionGroup']):
    obj: CollectionGroup

    factory = CollectionGroupFactoryService()



from __future__ import annotations

from django.urls import include, path


app_name = 'notification'

urlpatterns = [
    path('app/', include('django_spire.notification.app.urls', namespace='app')),
    path('sms/', include('django_spire.notification.sms.urls', namespace='sms')),
]



from __future__ import annotations

import uuid

from django.utils.timezone import now, timedelta

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.models import Notification
from django_spire.notification.sms.choices import SmsMediaContentTypeChoices
from django_spire.notification.sms.models import SmsTemporaryMedia
from django_spire.notification.sms.tests.factories import create_test_sms_notification


class SmsNotificationModelTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.sms_notification = create_test_sms_notification()

    def test_str(self):
        expected = f'{self.sms_notification.to_phone_number} - {self.sms_notification.notification.title}'
        assert str(self.sms_notification) == expected

    def test_notification_relationship(self):
        assert self.sms_notification.notification is not None
        assert isinstance(self.sms_notification.notification, Notification)

    def test_to_phone_number(self):
        assert self.sms_notification.to_phone_number == '5551234567'

    def test_media_url_default_none(self):
        assert self.sms_notification.media_url is None

    def test_temporary_media_default_none(self):
        assert self.sms_notification.temporary_media is None

    def test_with_media_url(self):
        sms_notification = create_test_sms_notification(
            media_url='https://example.com/image.png'
        )
        assert sms_notification.media_url == 'https://example.com/image.png'


class SmsTemporaryMediaModelTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.temporary_media = SmsTemporaryMedia.objects.create(
            content='base64encodedcontent',
            content_type=SmsMediaContentTypeChoices.PNG,
            name='test_image.png',
            expire_datetime=now() + timedelta(hours=1),
            external_access_key=uuid.uuid4(),
        )

    def test_str(self):
        expected = f'{self.temporary_media.name} - {self.temporary_media.content_type}'
        assert str(self.temporary_media) == expected

    def test_is_expired_false(self):
        assert self.temporary_media.is_expired() is False

    def test_is_expired_true(self):
        self.temporary_media.expire_datetime = now() - timedelta(hours=1)
        self.temporary_media.save()
        assert self.temporary_media.is_expired() is True

    def test_external_url(self):
        url = self.temporary_media.external_url
        assert str(self.temporary_media.external_access_key) in url

    def test_content_type_png(self):
        assert self.temporary_media.content_type == SmsMediaContentTypeChoices.PNG

    def test_content_type_jpeg(self):
        media = SmsTemporaryMedia.objects.create(
            content='base64encodedcontent',
            content_type=SmsMediaContentTypeChoices.JPEG,
            name='test_image.jpg',
            expire_datetime=now() + timedelta(hours=1),
            external_access_key=uuid.uuid4(),
        )
        assert media.content_type == SmsMediaContentTypeChoices.JPEG



from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib import messages
from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse

from django_spire.core.redirect import safe_redirect_url
from django_spire.contrib.form.confirmation_forms import DeleteConfirmationForm
from django_spire.contrib.form.utils import show_form_errors

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from django.db.models import Model


def dispatch_modal_delete_form_content(
    request: WSGIRequest,
    *,
    obj: Model,
    form_action: str,
    context_data: dict | None = None,
    activity_func: callable | None = None,
    auto_add_activity: bool = True,
    delete_func: callable | None = None,
    # Present and past tense of verb
    verbs: tuple[str, str] = ('delete', 'deleted'),
    return_url: str | None = None,
    template: str = 'django_spire/modal/content/dispatch_modal_delete_confirmation_content.html',
    show_success_message: bool = False
) -> HttpResponseRedirect | TemplateResponse:
    if context_data is None:
        context_data = {}

    model_name = obj._meta.model._meta.verbose_name

    if request.method == 'POST':
        form = DeleteConfirmationForm(request.POST, obj=obj)

        if form.is_valid():
            form.save(
                user=request.user,
                verbs=verbs,
                delete_func=delete_func,
                activity_func=activity_func,
                auto_add_activity=auto_add_activity
            )

            if show_success_message:
                messages.success(request, f'Successfully deleted {model_name}.')

        else:
            show_form_errors(request, form)

        if return_url is None:
            return_url = safe_redirect_url(request)

        return HttpResponseRedirect(return_url)

    base_context_data = {
        'form_title': f'{model_name} - {verbs[0].title()} {obj}',
        'form_action': form_action,
        'form_description': f'Are you sure you would like to {verbs[0].lower()} {model_name} "{obj}"?',
    }

    context_data = {**base_context_data, **context_data}

    return TemplateResponse(
        request,
        template=template,
        context=context_data
    )



from __future__ import annotations

from django.urls import include, path


app_name = 'app'

urlpatterns = [
    path('json/', include('django_spire.notification.app.urls.json_urls', namespace='json')),
    path('page/', include('django_spire.notification.app.urls.page_urls', namespace='page')),
    path('template/', include('django_spire.notification.app.urls.template_urls', namespace='template')),
]



from __future__ import annotations

from typing import TYPE_CHECKING

from dandy import Bot, Prompt

from django_spire.knowledge.intelligence.intel.entry_intel import EntriesIntel

if TYPE_CHECKING:
    from django_spire.knowledge.entry.models import Entry


class KnowledgeEntriesBot(Bot):
    llm_role = 'Knowledge Entry Search Assistant'
    llm_task = 'Read through the knowledge entries and return the most relevant for the user request.'
    llm_guidelines = (
        Prompt()
        .list([
            'Return 1 to 5 of the most relevant knowledge entries using block ids.',
            'The relevant headings should be the nearest heading above the selected knowledge entry block.',
            'Make sure the relevant heading text is from a heading with mark down formatting.',
            'When returning the relevant heading remove any of the markdown formating characters.',
        ])
    )
    llm_intel_class = EntriesIntel

    def process(self, user_input: str, entries: list[Entry]) -> EntriesIntel:
        entry_prompt = Prompt()
        entry_prompt.sub_heading('User Request')
        entry_prompt.line_break()
        entry_prompt.text(f'{user_input}')
        entry_prompt.line_break()
        entry_prompt.sub_heading('Knowledge Entries')
        entry_prompt.line_break()

        for entry in entries:
            entry_prompt.text(f'Entry: {entry.name}')

            for version_block in entry.current_version.blocks.all():
                if version_block.render_to_text() != '\n':
                    entry_prompt.text(f'{version_block.id}: {version_block.render_to_text()}')

            entry_prompt.line_break()

        return self.llm.prompt_to_intel(
            prompt=entry_prompt,
        )



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.knowledge.entry.version.block.choices import BlockTypeChoices
from django_spire.knowledge.entry.version.block.data.heading_data import HeadingEditorBlockData
from django_spire.knowledge.entry.version.block.data.list.data import ListEditorBlockData
from django_spire.knowledge.entry.version.block.data.text_data import TextEditorBlockData
from django_spire.knowledge.entry.version.block.tests.factories import create_test_version_block


class EntryVersionBlockModelTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.version_block = create_test_version_block()

    def test_editor_js_block_data_getter_text(self):
        self.version_block.type = BlockTypeChoices.TEXT
        self.version_block._block_data = {'text': 'Hello'}
        result = self.version_block.editor_js_block_data
        assert isinstance(result, TextEditorBlockData)
        assert result.text == 'Hello'

    def test_editor_js_block_data_getter_heading(self):
        self.version_block.type = BlockTypeChoices.HEADING
        self.version_block._block_data = {'text': 'Title', 'level': 1}
        result = self.version_block.editor_js_block_data
        assert isinstance(result, HeadingEditorBlockData)
        assert result.text == 'Title'
        assert result.level == 1

    def test_editor_js_block_data_getter_list(self):
        self.version_block.type = BlockTypeChoices.LIST
        self.version_block._block_data = {
            'style': 'unordered',
            'items': [{'content': 'Item 1', 'items': []}]
        }
        result = self.version_block.editor_js_block_data
        assert isinstance(result, ListEditorBlockData)

    def test_editor_js_block_data_setter(self):
        text_data = TextEditorBlockData(text='New text')
        self.version_block.editor_js_block_data = text_data
        assert self.version_block._block_data == {'text': 'New text'}
        assert 'New text' in self.version_block._text_data

    def test_update_editor_js_block_data_from_dict(self):
        self.version_block.type = BlockTypeChoices.TEXT
        self.version_block.update_editor_js_block_data_from_dict({'text': 'Updated'})
        assert self.version_block._block_data == {'text': 'Updated'}

    def test_render_to_text(self):
        self.version_block.type = BlockTypeChoices.TEXT
        self.version_block._block_data = {'text': 'Hello World'}
        result = self.version_block.render_to_text()
        assert 'Hello World' in result



from __future__ import annotations

from dandy import BaseIntel, BaseListIntel

from django_spire.knowledge.entry.models import Entry


class EntryIntel(BaseIntel):
    relevant_heading_text: str
    relevant_block_id: int

    def __str__(self):
        return self.relevant_heading_text

    @property
    def collection(self):
        return self.entry.collection

    @property
    def entry(self):
        return Entry.objects.get_by_version_block_id(self.relevant_block_id)


class EntriesIntel(BaseListIntel[EntryIntel]):
    entry_intel_list: list[EntryIntel]

    @property
    def unique(self) -> list[EntryIntel]:
        seen = set()
        result = []

        for entry_intel in self:
            if entry_intel.relevant_heading_text not in seen:
                seen.add(entry_intel.relevant_heading_text)
                result.append(entry_intel)

        return result



from __future__ import annotations

from typing import TYPE_CHECKING

from dandy import Prompt

if TYPE_CHECKING:
    from django.db.models import Model


def objective_prompt(
    model_class: type[Model],
    model_description: str,
    sector_description: str,
) -> Prompt:
    model_name = model_class._meta.verbose_name_plural.title()

    return (
        Prompt()
        .text(f'You are generating "{model_name}" that are described as "{model_description}".')
        .text(f'These {model_name} represent entities in our "{sector_description}" system.')
    )



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.sms.consts import (
    TWILIO_SMS_BATCH_SIZE_NAME,
    TWILIO_UNSUCCESSFUL_STATUSES,
)


class SmsConstsTests(BaseTestCase):
    def test_twilio_unsuccessful_statuses(self):
        assert 'failed' in TWILIO_UNSUCCESSFUL_STATUSES
        assert 'undelivered' in TWILIO_UNSUCCESSFUL_STATUSES
        assert len(TWILIO_UNSUCCESSFUL_STATUSES) == 2

    def test_twilio_sms_batch_size_name(self):
        assert TWILIO_SMS_BATCH_SIZE_NAME == 'TWILIO_SMS_BATCH_SIZE'



from __future__ import annotations

from unittest.mock import Mock, patch

from dandy.llm.request.message import MessageHistory
from django.test import RequestFactory, override_settings

from django_spire.ai.chat.message_intel import DefaultMessageIntel
from django_spire.ai.chat.router import SpireChatRouter
from django_spire.core.tests.test_cases import BaseTestCase


class TestSpireChatRouter(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.factory = RequestFactory()
        self.request = self.factory.get('/')
        self.request.user = self.super_user

    def test_router_can_be_instantiated(self) -> None:
        router = SpireChatRouter()

        assert isinstance(router, SpireChatRouter)

    def test_default_chat_callable_returns_default_without_permission(self) -> None:
        router = SpireChatRouter()

        self.request.user = Mock()
        self.request.user.has_perm.return_value = False

        result = router._default_chat_callable(
            request=self.request,
            user_input='Hello',
            message_history=None
        )

        assert isinstance(result, DefaultMessageIntel)
        assert result.text == 'Sorry, I could not find any information on that.'

    def test_default_chat_callable_calls_knowledge_workflow_with_permission(self) -> None:
        router = SpireChatRouter()

        with patch('django_spire.knowledge.intelligence.workflows.knowledge_workflow.knowledge_search_workflow') as mock_workflow:
            mock_workflow.return_value = DefaultMessageIntel(text='Knowledge response')

            result = router._default_chat_callable(
                request=self.request,
                user_input='Hello',
                message_history=None
            )

            mock_workflow.assert_called_once_with(
                request=self.request,
                user_input='Hello',
                message_history=None
            )

            assert isinstance(result, DefaultMessageIntel)

    def test_workflow_uses_intent_decoder(self) -> None:
        router = SpireChatRouter()

        with patch('django_spire.ai.chat.router.generate_intent_decoder') as mock_decoder:
            mock_decoder_instance = Mock()
            mock_decoder.return_value = mock_decoder_instance
            mock_decoder_instance.process.return_value = [lambda **kwargs: DefaultMessageIntel(text='Response')]

            result = router.workflow(
                request=self.request,
                user_input='Hello',
                message_history=None
            )

            mock_decoder.assert_called_once()
            assert isinstance(result, DefaultMessageIntel)

    @override_settings(DJANGO_SPIRE_AI_INTENT_CHAT_ROUTERS={})
    def test_workflow_with_no_intent_routers(self) -> None:
        router = SpireChatRouter()

        with patch.object(router, '_default_chat_callable') as mock_default:
            mock_default.return_value = DefaultMessageIntel(text='Default response')

            with patch('django_spire.ai.chat.router.generate_intent_decoder') as mock_decoder:
                mock_decoder_instance = Mock()
                mock_decoder.return_value = mock_decoder_instance
                mock_decoder_instance.process.return_value = [mock_default]

                result = router.workflow(
                    request=self.request,
                    user_input='Hello',
                    message_history=None
                )

                assert isinstance(result, DefaultMessageIntel)

    def test_workflow_passes_message_history_to_decoder(self) -> None:
        router = SpireChatRouter()
        message_history = MessageHistory()

        with patch('django_spire.ai.chat.router.generate_intent_decoder') as mock_decoder:
            mock_decoder_instance = Mock()
            mock_decoder.return_value = mock_decoder_instance
            mock_decoder_instance.process.return_value = [
                lambda **kwargs: DefaultMessageIntel(text='Response')
            ]

            router.workflow(
                request=self.request,
                user_input='Hello',
                message_history=message_history
            )

            mock_decoder.assert_called_once()

    def test_workflow_returns_message_intel(self) -> None:
        router = SpireChatRouter()

        with patch('django_spire.ai.chat.router.generate_intent_decoder') as mock_decoder:
            mock_decoder_instance = Mock()
            mock_decoder.return_value = mock_decoder_instance
            mock_decoder_instance.process.return_value = [
                lambda **kwargs: DefaultMessageIntel(text='Test Response')
            ]

            result = router.workflow(
                request=self.request,
                user_input='Hello',
                message_history=None
            )

            assert isinstance(result, DefaultMessageIntel)
            assert result.text == 'Test Response'



from __future__ import annotations

from django.urls import path

from django_spire.metric.report.views import page_views


app_name = 'report'

urlpatterns = [
    path('',
         page_views.report_view,
         name='report'),
]



from __future__ import annotations

from django.test import override_settings, RequestFactory

from django_spire.core.redirect.safe_redirect import (
    is_url_valid_and_safe,
    resolve_url,
    safe_redirect_url
)
from django_spire.core.tests.test_cases import BaseTestCase


class UrlsTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_is_url_valid_and_safe_empty_url(self) -> None:
        assert is_url_valid_and_safe('', {'localhost'}) is False

    def test_is_url_valid_and_safe_invalid_host(self) -> None:
        url = 'http://invalid.com'

        assert is_url_valid_and_safe(url, {'example.com'}) is False

    def test_is_url_valid_and_safe_invalid_scheme(self) -> None:
        url = 'ftp://example.com'

        assert is_url_valid_and_safe(url, {'example.com'}) is False

    def test_is_url_valid_and_safe_valid_url(self) -> None:
        url = 'http://example.com'

        assert is_url_valid_and_safe(url, {'example.com'}) is True

    def test_resolve_invalid_url(self) -> None:
        invalid_url = 'invalid:url'
        resolved_url = resolve_url(invalid_url)
        assert resolved_url == invalid_url


@override_settings(ALLOWED_HOSTS=['example.com'])
class ReturnUrlTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.request_factory = RequestFactory()

    def test_safe_redirect_url_invalid_return_url(self) -> None:
        request = self.request_factory.get(
            '/some_path',
            {'return_url': 'http://invalid.com'}
        )

        request.META['HTTP_HOST'] = 'example.com'
        response = safe_redirect_url(request)
        assert response == '/'

    def test_safe_redirect_url_return_url_with_encoded_characters(self) -> None:
        request = self.request_factory.get(
            '/some_path',
            {'return_url': 'http://example.com/valid%20path'}
        )

        request.META['HTTP_HOST'] = 'example.com'
        response = safe_redirect_url(request)
        assert response == 'http://example.com/valid%20path'

    def test_safe_redirect_url_return_url_with_invalid_characters(self) -> None:
        request = self.request_factory.get(
            '/some_path',
            {'return_url': 'http://example.com/invalid<>chars'}
        )

        request.META['HTTP_HOST'] = 'example.com'
        response = safe_redirect_url(request)
        assert response == '/'

    def test_safe_redirect_url_return_url_with_javascript_code(self) -> None:
        request = self.request_factory.get(
            '/some_path',
            {'return_url': 'javascript:alert("XSS")'}
        )

        request.META['HTTP_HOST'] = 'example.com'
        response = safe_redirect_url(request)
        assert response == '/'

    def test_safe_redirect_url_return_url_with_utf8_characters(self) -> None:
        request = self.request_factory.get(
            '/some_path',
            {'return_url': 'http://example.com/ümlaut'}
        )

        request.META['HTTP_HOST'] = 'example.com'
        response = safe_redirect_url(request)
        assert response == 'http://example.com/ümlaut'

    def test_safe_redirect_url_valid_return_url(self) -> None:
        request = self.request_factory.get(
            '/some_path',
            {'return_url': 'http://example.com/valid'}
        )

        request.META['HTTP_HOST'] = 'example.com'
        response = safe_redirect_url(request)
        assert response == 'http://example.com/valid'


@override_settings(ALLOWED_HOSTS=['example.com'])
class RefererTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.request_factory = RequestFactory()

    def test_safe_redirect_url_invalid_referer(self) -> None:
        request = self.request_factory.get('/some_path')
        request.META['HTTP_HOST'] = 'example.com'
        request.META['HTTP_REFERER'] = 'http://invalid.com'

        response = safe_redirect_url(request)
        assert response == '/'

    def test_safe_redirect_url_invalid_return_url(self) -> None:
        request = self.request_factory.get(
            '/some_path',
            {'return_url': 'http://invalid.com'}
        )

        request.META['HTTP_HOST'] = 'example.com'
        response = safe_redirect_url(request)
        assert response == '/'

    def test_safe_redirect_url_no_return_url_or_referer(self) -> None:
        request = self.request_factory.get('/some_path')
        request.META['HTTP_HOST'] = 'example.com'

        response = safe_redirect_url(request)
        assert response == '/'

    def test_safe_redirect_url_referer_with_encoded_characters(self) -> None:
        request = self.request_factory.get('/some_path')
        request.META['HTTP_HOST'] = 'example.com'
        request.META['HTTP_REFERER'] = 'http://example.com/valid%20path'

        response = safe_redirect_url(request)
        assert response == 'http://example.com/valid%20path'

    def test_safe_redirect_url_referer_with_invalid_characters(self) -> None:
        request = self.request_factory.get('/some_path')
        request.META['HTTP_HOST'] = 'example.com'
        request.META['HTTP_REFERER'] = 'http://example.com/invalid<>chars'

        response = safe_redirect_url(request)
        assert response == '/'

    def test_safe_redirect_url_referer_with_javascript_code(self) -> None:
        request = self.request_factory.get('/some_path')
        request.META['HTTP_HOST'] = 'example.com'
        request.META['HTTP_REFERER'] = 'javascript:alert("XSS")'

        response = safe_redirect_url(request)
        assert response == '/'

    def test_safe_redirect_url_referer_with_utf8_characters(self) -> None:
        request = self.request_factory.get('/some_path')
        request.META['HTTP_HOST'] = 'example.com'
        request.META['HTTP_REFERER'] = 'http://example.com/ümlaut'

        response = safe_redirect_url(request)
        assert response == 'http://example.com/ümlaut'

    def test_safe_redirect_url_valid_referer(self) -> None:
        request = self.request_factory.get('/some_path')
        request.META['HTTP_HOST'] = 'example.com'
        request.META['HTTP_REFERER'] = 'http://example.com/valid'

        response = safe_redirect_url(request)
        assert response == 'http://example.com/valid'

    def test_safe_redirect_url_valid_return_url(self) -> None:
        request = self.request_factory.get(
            '/some_path',
            {'return_url': 'http://example.com/valid'}
        )

        request.META['HTTP_HOST'] = 'example.com'
        response = safe_redirect_url(request)
        assert response == 'http://example.com/valid'



from __future__ import annotations

from django_spire.contrib.service import BaseDjangoModelService

from typing import TYPE_CHECKING

from django_spire.knowledge.entry.version.block.data.maps import EDITOR_JS_BLOCK_DATA_MAP
from django_spire.knowledge.entry.version.maps import FILE_TYPE_CONVERTER_MAP

if TYPE_CHECKING:
    from django_spire.file.models import File
    from django_spire.knowledge.entry.version.block.choices import BlockTypeChoices
    from django_spire.knowledge.entry.version.block.models import EntryVersionBlock
    from django_spire.knowledge.entry.version.models import EntryVersion


class EntryVersionBlockFactoryService(BaseDjangoModelService['EntryVersionBlock']):
    obj: EntryVersionBlock

    def create_blocks_from_file(
            self,
            file: File,
            entry_version: EntryVersion
    ) -> list[EntryVersionBlock]:
        if file.type not in FILE_TYPE_CONVERTER_MAP:
            return []

        converter = FILE_TYPE_CONVERTER_MAP[file.type](entry_version=entry_version)

        return self.obj_class.objects.bulk_create(
            converter.convert_file_to_blocks(file=file)
        )

    def create_validated_block(
            self,
            entry_version: EntryVersion,
            type: BlockTypeChoices,
            data: dict,
            order: int,
            tunes: dict = {},
            **kwargs,
    ):
        self.obj.version = entry_version
        self.obj.type = type
        self.obj.order = order
        self.obj.tunes = tunes
        self.obj.editor_js_block_data = EDITOR_JS_BLOCK_DATA_MAP[type](**data)

        self.obj.clean()

        return self.obj



from __future__ import annotations

import pytest

from django.test import TestCase

from django_spire.exceptions import DjangoSpireConfigurationError, DjangoSpireError


class TestDjangoSpireError(TestCase):
    def test_can_be_raised(self) -> None:
        with pytest.raises(DjangoSpireError):
            message = 'Test error'
            raise DjangoSpireError(message)

    def test_is_exception_subclass(self) -> None:
        assert issubclass(DjangoSpireError, Exception)

    def test_message(self) -> None:
        error = DjangoSpireError('Test error message')

        assert str(error) == 'Test error message'


class TestDjangoSpireConfigurationError(TestCase):
    def test_can_be_raised(self) -> None:
        with pytest.raises(DjangoSpireConfigurationError):
            message = 'Test error'
            raise DjangoSpireConfigurationError(message)

    def test_is_django_spire_error_subclass(self) -> None:
        assert issubclass(DjangoSpireConfigurationError, DjangoSpireError)

    def test_message(self) -> None:
        error = DjangoSpireConfigurationError('Configuration error message')

        assert str(error) == 'Configuration error message'
